# Licensed under a 3-clause BSD style license - see LICENSE.rst
import collections.abc
import copy
import html
import inspect
import itertools
import logging
import warnings
from itertools import zip_longest
import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.time import Time
from astropy.units import Quantity
from astropy.utils import lazyproperty
import matplotlib.pyplot as plt
from gammapy.irf import FoVAlignment
from gammapy.utils.coordinates import FoVAltAzFrame, FoVICRSFrame
from gammapy.utils.deprecation import GammapyDeprecationWarning
from gammapy.utils.fits import LazyFitsData, earth_location_to_dict
from gammapy.utils.metadata import CreatorMetaData, TargetMetaData, TimeInfoMetaData
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import Checker
from gammapy.utils.time import time_ref_to_dict, time_relative_to_ref
from .event_list import EventList, EventListChecker
from .filters import ObservationFilter
from .gti import GTI
from .metadata import ObservationMetaData
from .pointing import FixedPointingInfo

__all__ = ["Observation", "Observations"]

log = logging.getLogger(__name__)


class Observation:
    """In-memory observation.

    Parameters
    ----------
    obs_id : int, optional
        Observation id. Default is None
    aeff : `~gammapy.irf.EffectiveAreaTable2D`, optional
        Effective area. Default is None.
    edisp : `~gammapy.irf.EnergyDispersion2D`, optional
        Energy dispersion. Default is None.
    psf : `~gammapy.irf.PSF3D`, optional
        Point spread function. Default is None.
    bkg : `~gammapy.irf.Background3D`, optional
        Background rate model. Default is None.
    rad_max : `~gammapy.irf.RadMax2D`, optional
        Only for point-like IRFs: RAD_MAX table (energy dependent RAD_MAX)
        For a fixed RAD_MAX, create a RadMax2D with a single bin. Default is None.
    gti : `~gammapy.data.GTI`, optional
        Table with GTI start and stop time. Default is None.
    events : `~gammapy.data.EventList`, optional
        Event list. Default is None.
    obs_filter : `ObservationFilter`, optional
        Observation filter. Default is None.
    pointing : `~gammapy.data.FixedPointingInfo`, optional
        Pointing information. Default is None.
    location : `~astropy.coordinates.EarthLocation`, optional
        Earth location of the observatory. Default is None.
    """

    aeff = LazyFitsData(cache=False)
    edisp = LazyFitsData(cache=False)
    psf = LazyFitsData(cache=False)
    _bkg = LazyFitsData(cache=False)
    _rad_max = LazyFitsData(cache=True)
    _events = LazyFitsData(cache=False)
    _gti = LazyFitsData(cache=True)
    _pointing = LazyFitsData(cache=True)
    _meta = LazyFitsData(cache=True)

    def __init__(
        self,
        obs_id=None,
        meta=None,
        gti=None,
        aeff=None,
        edisp=None,
        psf=None,
        bkg=None,
        rad_max=None,
        events=None,
        obs_filter=None,
        pointing=None,
        location=None,
    ):
        self.obs_id = obs_id
        self.aeff = aeff
        self.edisp = edisp
        self.psf = psf
        self._bkg = bkg
        self._rad_max = rad_max
        self._gti = gti
        self._events = events
        self._pointing = pointing
        self._location = location  # this is part of the meta or is it data?
        self.obs_filter = obs_filter or ObservationFilter()
        self._meta = meta or ObservationMetaData()

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    @property
    def bkg(self):
        """Background of the observation."""
        from gammapy.irf import FoVAlignment

        bkg = self._bkg
        # used for backward compatibility of old HESS data
        try:
            if (
                bkg
                and self._meta
                and self._meta.optional
                and self._meta.optional["CREATOR"] == "SASH FITS::EventListWriter"
                and self._meta.optional["HDUVERS"] == "0.2"
            ):
                bkg._fov_alignment = FoVAlignment.REVERSE_LON_RADEC
        except KeyError:
            pass
        return bkg

    @bkg.setter
    def bkg(self, value):
        self._bkg = value

    @property
    def meta(self):
        """Return metadata container."""
        return self._meta

    @property
    def rad_max(self):
        """Rad max IRF. None if not available."""
        # prevent circular import
        from gammapy.irf import RadMax2D

        if self._rad_max is not None:
            return self._rad_max

        # load once to avoid trigger lazy loading it three times
        aeff = self.aeff
        if aeff is not None and aeff.is_pointlike:
            self._rad_max = RadMax2D.from_irf(aeff)
            return self._rad_max

        edisp = self.edisp
        if edisp is not None and edisp.is_pointlike:
            self._rad_max = RadMax2D.from_irf(self.edisp)

        return self._rad_max

    @property
    def available_hdus(self):
        """Which HDUs are available."""
        available_hdus = []
        keys = ["_events", "_gti", "aeff", "edisp", "psf", "_bkg", "_rad_max"]
        hdus = ["events", "gti", "aeff", "edisp", "psf", "bkg", "rad_max"]
        for key, hdu in zip(keys, hdus):
            available = self.__dict__.get(key, False)
            available_hdu = self.__dict__.get(f"_{hdu}_hdu", False)
            available_hdu_ = self.__dict__.get(f"_{key}_hdu", False)
            if available or available_hdu or available_hdu_:
                available_hdus.append(hdu)
        return available_hdus

    @property
    def available_irfs(self):
        """Which IRFs are available."""
        return [_ for _ in self.available_hdus if _ not in ["events", "gti"]]

    @property
    def events(self):
        """Event list of the observation as an `~gammapy.data.EventList`."""
        events = self.obs_filter.filter_events(self._events)
        return events

    @events.setter
    def events(self, value):
        if not isinstance(value, EventList):
            raise TypeError(f"events must be an EventList instance, got: {type(value)}")
        self._events = value

    @property
    def gti(self):
        """GTI of the observation as a `~gammapy.data.GTI`."""
        gti = self.obs_filter.filter_gti(self._gti)
        return gti

    @gti.setter
    def gti(self, value):
        if self.gti is not None and self.events is not None:
            raise ValueError("GTI cannot be redefined if events is not None.")
        if not isinstance(value, GTI):
            raise TypeError(f"GTI must be an GTI instance, got: {type(value)}")
        self._gti = value

    @staticmethod
    def _get_obs_info(
        pointing, deadtime_fraction, time_start, time_stop, reference_time, location
    ):
        """Create observation information dictionary from in memory data."""
        obs_info = {
            "DEADC": 1 - deadtime_fraction,
        }
        if isinstance(pointing, SkyCoord):
            obs_info["RA_PNT"] = pointing.icrs.ra.deg
            obs_info["DEC_PNT"] = pointing.icrs.dec.deg

        obs_info.update(time_ref_to_dict(reference_time))
        obs_info["TSTART"] = time_relative_to_ref(time_start, obs_info).to_value(u.s)
        obs_info["TSTOP"] = time_relative_to_ref(time_stop, obs_info).to_value(u.s)

        if location is not None:
            obs_info.update(earth_location_to_dict(location))

        return obs_info

    @classmethod
    def create(
        cls,
        pointing,
        location=None,
        obs_id=0,
        livetime=None,
        tstart=None,
        tstop=None,
        irfs=None,
        deadtime_fraction=0.0,
        reference_time=Time("2000-01-01 00:00:00"),
    ):
        """Create an observation.

        User must either provide the livetime, or the start and stop times.

        Parameters
        ----------
        pointing : `~gammapy.data.FixedPointingInfo` or `~astropy.coordinates.SkyCoord`
            Pointing information.
        location : `~astropy.coordinates.EarthLocation`, optional
            Earth location of the observatory. Default is None.
        obs_id : int, optional
            Observation ID as identifier. Default is 0.
        livetime : ~astropy.units.Quantity`, optional
            Livetime exposure of the simulated observation. Default is None.
        tstart : `~astropy.time.Time` or `~astropy.units.Quantity`, optional
            Start time of observation as `~astropy.time.Time` or duration
            relative to `reference_time`. Default is None.
        tstop : `astropy.time.Time` or `~astropy.units.Quantity`, optional
            Stop time of observation as `~astropy.time.Time` or duration
            relative to `reference_time`. Default is None.
        irfs : dict, optional
            IRFs used for simulating the observation: `bkg`, `aeff`, `psf`, `edisp`, `rad_max`. Default is None.
        deadtime_fraction : float, optional
            Deadtime fraction. Default is 0.
        reference_time : `~astropy.time.Time`, optional
            the reference time to use in GTI definition. Default is `~astropy.time.Time("2000-01-01 00:00:00")`.

        Returns
        -------
        obs : `gammapy.data.MemoryObservation`
            Observation.
        """
        if tstart is None:
            tstart = reference_time.copy()

        if tstop is None:
            tstop = tstart + Quantity(livetime)

        gti = GTI.create(tstart, tstop, reference_time=reference_time)
        obs_info = cls._get_obs_info(
            pointing=pointing,
            deadtime_fraction=deadtime_fraction,
            time_start=gti.time_start[0],
            time_stop=gti.time_stop[0],
            reference_time=reference_time,
            location=location,
        )

        time_info = TimeInfoMetaData(
            time_start=gti.time_start[0],
            time_stop=gti.time_stop[-1],
            reference_time=reference_time,
        )

        meta = ObservationMetaData(
            deadtime_fraction=deadtime_fraction,
            location=location,
            time_info=time_info,
            creation=CreatorMetaData(),
            target=TargetMetaData(),
        )

        if not isinstance(pointing, FixedPointingInfo):
            warnings.warn(
                "Pointing will be required to be provided as FixedPointingInfo",
                GammapyDeprecationWarning,
            )
            pointing = FixedPointingInfo.from_fits_header(obs_info)

        return cls(
            obs_id=obs_id,
            meta=meta,
            gti=gti,
            aeff=irfs.get("aeff"),
            bkg=irfs.get("bkg"),
            edisp=irfs.get("edisp"),
            psf=irfs.get("psf"),
            rad_max=irfs.get("rad_max"),
            pointing=pointing,
            location=location,
        )

    @property
    def tstart(self):
        """Observation start time as a `~astropy.time.Time` object."""
        return self.gti.time_start[0]

    @property
    def tstop(self):
        """Observation stop time as a `~astropy.time.Time` object."""
        return self.gti.time_stop[0]

    @property
    def tmid(self):
        """Midpoint between start and stop time as a `~astropy.time.Time` object."""
        return self.tstart + 0.5 * (self.tstop - self.tstart)

    @property
    def observation_time_duration(self):
        """Observation time duration in seconds as a `~astropy.units.Quantity`.

        The wall time, including dead-time.
        """
        return self.gti.time_sum

    @property
    def observation_live_time_duration(self):
        """Live-time duration in seconds as a `~astropy.units.Quantity`.

        The dead-time-corrected observation time.

        Computed as ``t_live = t_observation * (1 - f_dead)``
        where ``f_dead`` is the dead-time fraction.
        """
        return (
            self.observation_time_duration
            * (1 - self.observation_dead_time_fraction)
            * self.obs_filter.livetime_fraction
        )

    @property
    def observation_dead_time_fraction(self):
        """Dead-time fraction (float).

        Defined as dead-time over observation time.

        Dead-time is defined as the time during the observation
        where the detector didn't record events:
        https://en.wikipedia.org/wiki/Dead_time
        https://ui.adsabs.harvard.edu/abs/2004APh....22..285F

        The dead-time fraction is used in the live-time computation,
        which in turn is used in the exposure and flux computation.
        """
        return self.meta.deadtime_fraction

    @property
    def pointing(self):
        """Get the pointing for the observation as a `~gammapy.data.FixedPointingInfo` object."""
        return self._pointing

    def get_pointing_altaz(self, time):
        """Get the pointing in alt-az for given time."""
        return self.pointing.get_altaz(time, self.observatory_earth_location)

    def get_pointing_icrs(self, time):
        """Get the pointing in ICRS for given time."""
        return self.pointing.get_icrs(time, self.observatory_earth_location)

    def _get_fov_altaz_frame(self, time):
        """Get the `~gammapy.utils.coordinates.FoVAltAzFrame` (FoV Frame aligned on AltAz) for given time."""
        origin = self.get_pointing_altaz(time)
        location = self.observatory_earth_location
        return FoVAltAzFrame(origin=origin, location=location, obstime=time)

    def _get_fov_icrs_frame(self, time):
        """Get the `~gammapy.utils.coordinates.FoVICRSFrame` (FoV Frame aligned on ICRS) for given time."""
        origin = self.get_pointing_icrs(time)
        return FoVICRSFrame(origin=origin)

    def get_fov_frame(self, time, alignment):
        """Get the `~gammapy.utils.coordinates.FoVICRSFrame` or `~gammapy.utils.coordinates.FoVAltAzFrame` for given time.

        Parameters
        ----------
        time : `~astropy.time.Time`
            Times at which to extract the frame.
        alignment : `~gammapy.irf.FoVAlignment`
            Alignment of the field-of-view frame.

        Returns
        -------
        fov_frame : `~gammapy.utils.coordinates.FoVICRSFrame` or `~gammapy.utils.coordinates.FoVAltAzFrame`
            The field-of-view frame with the required alignment (on ICRS or AltAz).
        """
        if alignment in [FoVAlignment.RADEC, FoVAlignment.REVERSE_LON_RADEC]:
            return self._get_fov_icrs_frame(time)
        elif alignment == FoVAlignment.ALTAZ:
            return self._get_fov_altaz_frame(time)
        else:
            raise ValueError(f"Unknown FoVAlignment {alignment}")

    @property
    def observatory_earth_location(self):
        """Observatory location as an `~astropy.coordinates.EarthLocation` object."""
        if self._location is None:
            return self.meta.location
        return self._location

    @lazyproperty
    def target_radec(self):
        """Target RA / DEC sky coordinates as a `~astropy.coordinates.SkyCoord` object."""
        return self.meta.target.position

    def __str__(self):
        pointing = self.get_pointing_icrs(self.tmid)
        ra = pointing.ra.deg
        dec = pointing.dec.deg

        pointing = f"{ra:.1f} deg, {dec:.1f} deg\n"
        # TODO: Which target was observed?
        # TODO: print info about available HDUs for this observation ...
        return (
            f"{self.__class__.__name__}\n\n"
            f"\tobs id            : {self.obs_id} \n "
            f"\ttstart            : {self.tstart.mjd:.2f}\n"
            f"\ttstop             : {self.tstop.mjd:.2f}\n"
            f"\tduration          : {self.observation_time_duration:.2f}\n"
            f"\tpointing (icrs)   : {pointing}\n"
            f"\tdeadtime fraction : {self.observation_dead_time_fraction:.1%}\n"
        )

    def check(self, checks="all"):
        """Run checks.

        This is a generator that yields a list of dictionary.
        """
        checker = ObservationChecker(self)
        return checker.run(checks=checks)

    def peek(self, figsize=(15, 10)):
        """Quick-look plots in a few panels.

        This method creates a figure displaying the available events and IRFs.
        For example:

        * Events 2D map : events sky map
        * Effective area 2D map : effective area as a function of FoV offset and true energy
        * Background rate 2D map : background rate as a function of FoV offset and energy
        * Energy dispersion 2D map : migration as a function of true energy for a given offset
        * Point spread function plot : containment radius as a function of energy for various
          containment fractions
        * Rad max plot : radius of the directional cut (rad max) as a function of energy


        Parameters
        ----------
        figsize : tuple, optional
            Figure size. Default is (15, 10).
        """
        plottable_hds = ["events", "aeff", "psf", "edisp", "bkg", "rad_max"]

        plot_hdus = list(set(plottable_hds) & set(self.available_hdus))
        plot_hdus.sort()

        n_irfs = len(plot_hdus)
        nrows = n_irfs // 2
        ncols = 2 + n_irfs % 2

        fig, axes = plt.subplots(
            nrows=nrows,
            ncols=ncols,
            figsize=figsize,
            gridspec_kw={"wspace": 0.3, "hspace": 0.3},
        )

        for idx, (ax, name) in enumerate(zip_longest(axes.flat, plot_hdus)):
            if name == "aeff":
                self.aeff.plot(ax=ax)
                ax.set_title("Effective area")

            if name == "bkg":
                bkg = self.bkg
                if not bkg.has_offset_axis:
                    bkg = bkg.to_2d()
                bkg.plot(ax=ax)
                ax.set_title("Background rate")

            if name == "psf":
                self.psf.plot_containment_radius_vs_energy(ax=ax)
                ax.set_title("Point spread function")

            if name == "edisp":
                self.edisp.plot_bias(ax=ax, add_cbar=True)
                ax.set_title("Energy dispersion")

            if name == "rad_max":
                self.rad_max.plot_rad_max_vs_energy(ax=ax)
                ax.set_title("Rad max")

            if name == "events":
                m = self.events._counts_image(allsky=False)
                ax.remove()
                ax = fig.add_subplot(nrows, ncols, idx + 1, projection=m.geom.wcs)
                m.plot(ax=ax, stretch="sqrt", vmin=0, add_cbar=True)
                ax.set_title("Events")

            if name is None:
                ax.set_visible(False)

    def select_time(self, time_interval):
        """Select a time interval of the observation.

        Parameters
        ----------
        time_interval : `astropy.time.Time`
            Start and stop time of the selected time interval.
            For now, we only support a single time interval.

        Returns
        -------
        new_obs : `~gammapy.data.Observation`
            A new observation instance of the specified time interval.
        """
        new_obs_filter = self.obs_filter.copy()
        new_obs_filter.time_filter = time_interval
        obs = copy.deepcopy(self)
        obs.obs_filter = new_obs_filter
        return obs

    @classmethod
    def read(cls, event_file, irf_file=None, checksum=False):
        """Create an Observation from a Event List and an (optional) IRF file.

        Parameters
        ----------
        event_file : str or `~pathlib.Path`
            Path to the FITS file containing the event list and the GTI.
        irf_file : str or `~pathlib.Path`, optional
            Path to the FITS file containing the IRF components. Default is None.
            If None, the IRFs will be read from the event file.
        checksum : bool
            If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.

        Returns
        -------
        observation : `~gammapy.data.Observation`
            Observation with the events and the IRF read from the file.
        """
        from gammapy.irf.io import load_irf_dict_from_file

        events = EventList.read(event_file, checksum=checksum)

        gti = GTI.read(event_file, checksum=checksum)

        irf_file = irf_file if irf_file is not None else event_file
        irf_dict = load_irf_dict_from_file(irf_file)

        obs_info = events.table.meta

        meta = ObservationMetaData.from_header(obs_info)
        return cls(
            events=events,
            gti=gti,
            obs_id=meta.obs_info.obs_id,
            pointing=FixedPointingInfo.from_fits_header(obs_info),
            meta=meta,
            location=meta.location,
            **irf_dict,
        )

    def write(
        self, path, overwrite=False, format="gadf", include_irfs=True, checksum=False
    ):
        """
        Write this observation into `~pathlib.Path` using the specified format.

        Parameters
        ----------
        path : str or `~pathlib.Path`
            Path for the output file.
        overwrite : bool, optional
            Overwrite existing file. Default is False.
        format : {"gadf"}
            Output format, currently only "gadf" is supported. Default is "gadf".
        include_irfs : bool, optional
            Whether to include irf components in the output file. Default is True.
        checksum : bool, optional
            When True adds both DATASUM and CHECKSUM cards to the headers written to the file.
            Default is False.
        """
        if format != "gadf":
            raise ValueError(f'Only the "gadf" format is supported, got {format}')

        path = make_path(path)

        primary = fits.PrimaryHDU()

        creation = self.meta.creation or CreatorMetaData()
        primary.header.update(creation.to_header(format))

        hdul = fits.HDUList([primary])

        events = self.events
        if events is not None:
            events_hdu = events.to_table_hdu(format=format)

            if self.pointing is None:
                raise ValueError(
                    "Cannot write Observation in gadf format: Missing pointing."
                )

            events_hdu.header.update(
                self.pointing.to_fits_header(time_ref=events.time_ref)
            )
            hdul.append(events_hdu)

        gti = self.gti
        if gti is not None:
            hdul.append(gti.to_table_hdu(format=format))

        if include_irfs:
            for irf_name in self.available_irfs:
                irf = getattr(self, irf_name)
                if irf is not None:
                    hdul.append(irf.to_table_hdu(format="gadf-dl3"))

        hdul.writeto(path, overwrite=overwrite, checksum=checksum)

    def copy(self, in_memory=False, **kwargs):
        """Copy observation.

        Overwriting `Observation` arguments requires the ``in_memory`` argument to be true.

        Parameters
        ----------
        in_memory : bool, optional
            Copy observation in memory. Default is False.
        **kwargs : dict, optional
            Keyword arguments passed to `Observation`.

        Examples
        --------
        >>> from gammapy.data import Observation
        >>> obs = Observation.read("$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz")
        >>> obs_copy = obs.copy(in_memory=True, obs_id=1234)
        >>> print(obs_copy) # doctest: +SKIP

        Returns
        -------
        obs : `Observation`
            Copied observation.
        """
        if in_memory:
            argnames = inspect.getfullargspec(self.__init__).args
            # TODO: remove once obs_info is removed from the list of arguments in __init__
            argnames.remove("self")

            for name in argnames:
                if name == "location":
                    attr = "observatory_earth_location"
                else:
                    attr = name
                value = getattr(self, attr)
                kwargs.setdefault(name, copy.deepcopy(value))
            return self.__class__(**kwargs)

        if kwargs:
            raise ValueError("Overwriting arguments requires to set 'in_memory=True'")

        return copy.deepcopy(self)


class Observations(collections.abc.MutableSequence):
    """Container class that holds a list of observations.

    Parameters
    ----------
    observations : list
        A list of `~gammapy.data.Observation`.
    """

    def __init__(self, observations=None):
        self._observations = []

        if observations is None:
            observations = []

        for obs in observations:
            self.append(obs)

    def __getitem__(self, item):
        if isinstance(item, (list, np.ndarray)) and all(
            isinstance(x, str) for x in item
        ):
            return self.__class__([self._observations[self.index(_)] for _ in item])
        elif isinstance(item, (slice, list, np.ndarray)):
            return self.__class__(list(np.array(self._observations)[item]))
        else:
            return self._observations[self.index(item)]

    def __delitem__(self, key):
        del self._observations[self.index(key)]

    def __setitem__(self, key, obs):
        if isinstance(obs, Observation):
            if obs in self:
                log.warning(
                    f"Observation with obs_id {obs.obs_id} already belongs to Observations."
                )
            self._observations[self.index(key)] = obs
        else:
            raise TypeError(f"Invalid type: {type(obs)!r}")

    def insert(self, idx, obs):
        if isinstance(obs, Observation):
            if obs in self:
                log.warning(
                    f"Observation with obs_id {obs.obs_id} already belongs to Observations."
                )
            self._observations.insert(idx, obs)
        else:
            raise TypeError(f"Invalid type: {type(obs)!r}")

    def __len__(self):
        return len(self._observations)

    def __str__(self):
        s = self.__class__.__name__ + "\n"
        s += "Number of observations: {}\n".format(len(self))
        for obs in self:
            s += str(obs)
        return s

    def index(self, key):
        if isinstance(key, (int, slice)):
            return key
        elif isinstance(key, str):
            return self.ids.index(key)
        elif isinstance(key, Observation):
            return self._observations.index(key)
        else:
            raise TypeError(f"Invalid type: {type(key)!r}")

    @property
    def ids(self):
        """List of observation IDs (`list`)."""
        return [str(obs.obs_id) for obs in self]

    def select_time(self, time_intervals):
        """Select a time interval of the observations.

        Parameters
        ----------
        time_intervals : `astropy.time.Time` or list of `astropy.time.Time`
            List of start and stop time of the time intervals or one time interval.

        Returns
        -------
        new_observations : `~gammapy.data.Observations`
            A new Observations instance of the specified time intervals.
        """
        new_obs_list = []
        if isinstance(time_intervals, Time):
            time_intervals = [time_intervals]

        for time_interval in time_intervals:
            for obs in self:
                if (obs.tstart < time_interval[1]) & (obs.tstop > time_interval[0]):
                    new_obs = obs.select_time(time_interval)
                    new_obs_list.append(new_obs)

        return self.__class__(new_obs_list)

    def _ipython_key_completions_(self):
        return self.ids

    def group_by_label(self, labels):
        """Split observations in multiple groups of observations.

        Parameters
        ----------
        labels : list or `numpy.ndarray`
            Array of group labels.

        Returns
        -------
        obs_clusters : dict of `~gammapy.data.Observations`
            Dictionary of Observations instance, one instance for each group.
        """
        obs_groups = {}
        for label in np.unique(labels):
            observations = self.__class__(
                [obs for k, obs in enumerate(self) if labels[k] == label]
            )
            obs_groups[f"group_{label}"] = observations
        return obs_groups

    @classmethod
    def from_stack(cls, observations_list):
        # TODO : Do more check when stacking observations when we have metadata.
        """Create a new `Observations` instance by concatenating a list of `Observations` objects.

        Parameters
        ----------
        observations_list : list of `~gammapy.data.Observations`
            The list of `Observations` to stack.

        Returns
        -------
        observations : `~gammapy.data.Observations`
            The `Observations` object resulting from stacking all the `Observations` in `observation_list`.
        """
        obs = itertools.chain(*observations_list)
        return cls(list(obs))

    def in_memory_generator(self):
        """Iterate over the observation and yield an in memory copy of the observation."""
        for obs in self:
            obs_copy = obs.copy(in_memory=True)
            yield obs_copy


class ObservationChecker(Checker):
    """Check an observation.

    Checks data format and a bit about the content.
    """

    CHECKS = {
        "events": "check_events",
        "gti": "check_gti",
        "aeff": "check_aeff",
        "edisp": "check_edisp",
        "psf": "check_psf",
    }

    def __init__(self, observation):
        self.observation = observation

    def _record(self, level="info", msg=None):
        return {"level": level, "obs_id": self.observation.obs_id, "msg": msg}

    def check_events(self):
        yield self._record(level="debug", msg="Starting events check")

        try:
            events = self.observation.events
        except Exception:
            yield self._record(level="warning", msg="Loading events failed")
            return

        yield from EventListChecker(events).run()

    # TODO: split this out into a GTIChecker
    def check_gti(self):
        yield self._record(level="debug", msg="Starting gti check")

        try:
            gti = self.observation.gti
        except Exception:
            yield self._record(level="warning", msg="Loading GTI failed")
            return

        if len(gti.table) == 0:
            yield self._record(level="error", msg="GTI table has zero rows")

        columns_required = ["START", "STOP"]
        for name in columns_required:
            if name not in gti.table.colnames:
                yield self._record(level="error", msg=f"Missing table column: {name!r}")

        # TODO: Check that header keywords agree with table entries
        # TSTART, TSTOP, MJDREFI, MJDREFF

        # Check that START and STOP times are consecutive
        # times = np.ravel(self.table['START'], self.table['STOP'])
        # # TODO: not sure this is correct ... add test with a multi-gti table from Fermi.
        # if not np.all(np.diff(times) >= 0):
        #     yield 'GTIs are not consecutive or sorted.'

    # TODO: add reference times for all instruments and check for this
    # Use TELESCOP header key to check which instrument it is.
    def _check_times(self):
        """Check if various times are consistent.

        The headers and tables of the FITS EVENTS and GTI extension
        contain various observation and event time information.
        """
        # http://fermi.gsfc.nasa.gov/ssc/data/analysis/documentation/Cicerone/Cicerone_Data/Time_in_ScienceTools.html
        # https://cchesswiki.in2p3.fr/hess/working_groups/analysis_and_reconstruction_working_group/ar_active_tasks/hess_fits_data/hess_fits_data__references_and_checks
        telescope_met_refs = {
            "FERMI": Time("2001-01-01T00:00:00"),
            "HESS": Time("2001-01-01T00:00:00"),
        }

        meta = self.dset.event_list.table.meta
        telescope = meta["TELESCOP"]

        if telescope in telescope_met_refs.keys():
            dt = self.time_ref - telescope_met_refs[telescope]
            if dt > self.accuracy["time"]:
                yield self._record(
                    level="error", msg="Reference time incorrect for telescope"
                )

    def check_aeff(self):
        yield self._record(level="debug", msg="Starting aeff check")

        try:
            aeff = self.observation.aeff
        except Exception:
            yield self._record(level="warning", msg="Loading aeff failed")
            return

        # Check that thresholds are meaningful for aeff
        if (
            "LO_THRES" in aeff.meta
            and "HI_THRES" in aeff.meta
            and aeff.meta["LO_THRES"] >= aeff.meta["HI_THRES"]
        ):
            yield self._record(
                level="error", msg="LO_THRES >= HI_THRES in effective area meta data"
            )

        # Check that data isn't all null
        if np.max(aeff.data.data) <= 0:
            yield self._record(
                level="error", msg="maximum entry of effective area is <= 0"
            )

    def check_edisp(self):
        yield self._record(level="debug", msg="Starting edisp check")

        try:
            edisp = self.observation.edisp
        except Exception:
            yield self._record(level="warning", msg="Loading edisp failed")
            return

        # Check that data isn't all null
        if np.max(edisp.data.data) <= 0:
            yield self._record(level="error", msg="maximum entry of edisp is <= 0")

    def check_psf(self):
        yield self._record(level="debug", msg="Starting psf check")

        try:
            self.observation.psf
        except Exception:
            yield self._record(level="warning", msg="Loading psf failed")
            return



"""
Morphological energy dependence estimation
==========================================

Learn how to test for energy-dependent morphology in your dataset.

Prerequisites
-------------
Knowledge on data reduction and datasets used in Gammapy, for example see
the :doc:`/tutorials/data/hess` and :doc:`/tutorials/analysis-2d/ring_background` tutorials.


Context
-------

This tutorial introduces a method to investigate the potential of energy-dependent morphology from spatial maps.
It is possible for gamma-ray sources to exhibit energy-dependent morphology, in which the spatial morphology of
the gamma rays varies across different energy bands. This is plausible for different source types, including pulsar
wind nebulae (PWNe) and supernova remnants. HESS J1825−137 is a well-known example of a PWNe which shows a clear
energy-dependent gamma-ray morphology (see `Aharonian et al., 2006 <https://ui.adsabs.harvard.edu/abs/2019A%26A...621A.116H/abstract>`__,
`H.E.S.S. Collaboration et al., 2019 <https://ui.adsabs.harvard.edu/abs/2006A%26A...460..365A/abstract>`__ and
`Principe et al., 2020 <https://ui.adsabs.harvard.edu/abs/2020A%26A...640A..76P/abstract>`__.)

Many different techniques of measuring this energy-dependence have been utilised over the years.
The method shown here is to perform a morphological fit of extension and position in various energy slices and
compare this with a global morphology fit.


**Objective: Perform an energy-dependent morphology study of a mock source.**


Tutorial overview
-----------------

This tutorial consists of two main steps.

Firstly, the user defines the initial `~gammapy.modeling.models.SkyModel` based on previous investigations
and selects the energy bands of interest to test for energy dependence. The null hypothesis is defined as
only the background component being free (norm). The alternative hypothesis introduces the source model.
The results of this first step show the significance of the source above the background in each energy band.

The second step is to quantify any energy-dependent morphology. The null hypothesis is determined by performing
a joint fit of the parameters. In the alternative hypothesis, the free parameters of the model are fit
individually within each energy band.


Setup
-----

"""

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
import matplotlib.pyplot as plt
from IPython.display import display
from gammapy.datasets import Datasets, MapDataset
from gammapy.estimators import EnergyDependentMorphologyEstimator
from gammapy.estimators.energydependentmorphology import weighted_chi2_parameter
from gammapy.maps import Map
from gammapy.modeling.models import (
    GaussianSpatialModel,
    PowerLawSpectralModel,
    SkyModel,
)
from gammapy.stats.utils import ts_to_sigma

######################################################################
# Obtain the data to use
# ----------------------
#
# Utilise the pre-defined dataset within `$GAMMAPY_DATA`.
#
# P.S.: do not forget to set up your environment variable `$GAMMAPY_DATA`
# to your local directory.

stacked_dataset = MapDataset.read(
    "$GAMMAPY_DATA/estimators/mock_DL4/dataset_energy_dependent.fits.gz"
)
datasets = Datasets([stacked_dataset])


######################################################################
# Define the energy edges of interest. These will be utilised to
# investigate the potential of energy-dependent morphology in the dataset.

energy_edges = [1, 3, 5, 20] * u.TeV


######################################################################
# Define the spectral and spatial models of interest. We utilise
# a `~gammapy.modeling.models.PowerLawSpectralModel` and a
# `~gammapy.modeling.models.GaussianSpatialModel` to test the energy-dependent
# morphology component in each energy band. A standard 3D fit (see the
# :doc:`/tutorials/analysis-3d/analysis_3d` tutorial)
# is performed, then the best fit model is utilised here for the initial parameters
# in each model.

source_position = SkyCoord(5.58, 0.2, unit="deg", frame="galactic")

spectral_model = PowerLawSpectralModel(
    index=2.5, amplitude=9.8e-12 * u.Unit("cm-2 s-1 TeV-1"), reference=1.0 * u.TeV
)

spatial_model = GaussianSpatialModel(
    lon_0=source_position.l,
    lat_0=source_position.b,
    frame="galactic",
    sigma=0.2 * u.deg,
)

# Limit the search for the position on the spatial model
spatial_model.lon_0.min = source_position.galactic.l.deg - 0.8
spatial_model.lon_0.max = source_position.galactic.l.deg + 0.8
spatial_model.lat_0.min = source_position.galactic.b.deg - 0.8
spatial_model.lat_0.max = source_position.galactic.b.deg + 0.8

model = SkyModel(spatial_model=spatial_model, spectral_model=spectral_model, name="src")

######################################################################
# Run Estimator
# -------------
#
# We can now run the energy-dependent estimation tool and explore the results.
#
# Let's start with the initial hypothesis, in which the source is introduced
# to compare with the background. We specify which parameters we
# wish to use for testing the energy dependence.
# To test for the energy dependence, it is recommended to keep the position and
# extension parameters free. This allows them to be used for fitting the spatial model
# in each energy band.
#

model.spatial_model.lon_0.frozen = False
model.spatial_model.lat_0.frozen = False
model.spatial_model.sigma.frozen = False

model.spectral_model.amplitude.frozen = False
model.spectral_model.index.frozen = True

datasets.models = model

estimator = EnergyDependentMorphologyEstimator(energy_edges=energy_edges, source="src")

######################################################################
# Show the results tables
# -----------------------
#
# The results of the source signal above the background in each energy bin
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Firstly, the estimator is run to produce the results.
# The table here shows the ∆(TS) value, the number of degrees of freedom (df)
# and the significance (σ) in each energy bin. The significance values here show that each
# energy band has significant signal above the background.
#

results = estimator.run(datasets)
table_bkg_src = Table(results["src_above_bkg"])
display(table_bkg_src)

######################################################################
# The results for testing energy dependence
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Next, the results of the energy-dependent estimator are shown.
# The table shows the various free parameters for the joint fit for :math:`H_0` across the entire
# energy band and for each energy bin shown for :math:`H_1`.

ts = results["energy_dependence"]["delta_ts"]
df = results["energy_dependence"]["df"]
sigma = ts_to_sigma(ts, df=df)

print(f"The delta_ts for the energy-dependent study: {ts:.3f}.")
print(f"Converting this to a significance gives: {sigma:.3f} \u03c3")

results_table = Table(results["energy_dependence"]["result"])
display(results_table)


######################################################################
# The chi-squared value for each parameter of interest
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# We can also utilise the `~gammapy.estimators.energydependence.weighted_chi2_parameter`
# function for each parameter.
#
# The weighted chi-squared significance for the ``sigma``, ``lat_0`` and ``lon_0`` values.
#

display(
    Table(
        weighted_chi2_parameter(
            results["energy_dependence"]["result"],
            parameters=["sigma", "lat_0", "lon_0"],
        )
    )
)

######################################################################
# Note: The chi-squared parameter does not include potential correlation between the
# parameters, so it should be used cautiously.
#


######################################################################
# Plotting the results
# --------------------

empty_map = Map.create(
    skydir=spatial_model.position, frame=spatial_model.frame, width=1, binsz=0.02
)

colors = ["red", "blue", "green", "magenta"]

fig = plt.figure(figsize=(6, 4))
ax = empty_map.plot(cmap="Greys")

lat_0 = results["energy_dependence"]["result"]["lat_0"][1:]
lat_0_err = results["energy_dependence"]["result"]["lat_0_err"][1:]
lon_0 = results["energy_dependence"]["result"]["lon_0"][1:]
lon_0_err = results["energy_dependence"]["result"]["lon_0_err"][1:]
sigma = results["energy_dependence"]["result"]["sigma"][1:]
sigma_err = results["energy_dependence"]["result"]["sigma_err"][1:]

for i in range(len(lat_0)):
    model_plot = GaussianSpatialModel(
        lat_0=lat_0[i], lon_0=lon_0[i], sigma=sigma[i], frame=spatial_model.frame
    )
    model_plot.lat_0.error = lat_0_err[i]
    model_plot.lon_0.error = lon_0_err[i]
    model_plot.sigma.error = sigma_err[i]

    model_plot.plot_error(
        ax=ax,
        which="all",
        kwargs_extension={"facecolor": colors[i], "edgecolor": colors[i]},
        kwargs_position={"color": colors[i]},
    )
plt.show()


# sphinx_gallery_thumbnail_number = 1



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import abc
import html
import inspect
from copy import deepcopy
import numpy as np
from astropy import units as u
from gammapy.maps import MapAxis
from gammapy.modeling.models import ModelBase

__all__ = ["Estimator"]


class Estimator(abc.ABC):
    """Abstract estimator base class."""

    _available_selection_optional = {}

    @property
    @abc.abstractmethod
    def tag(self):
        pass

    @abc.abstractmethod
    def run(self, datasets):
        pass

    @property
    def selection_optional(self):
        """Set optional selection."""
        return self._selection_optional

    @selection_optional.setter
    def selection_optional(self, selection):
        """Set optional selection."""
        available = self._available_selection_optional

        if selection is None:
            self._selection_optional = []
        elif "all" in selection:
            self._selection_optional = available
        else:
            if set(selection).issubset(set(available)):
                self._selection_optional = selection
            else:
                difference = set(selection).difference(set(available))
                raise ValueError(f"{difference} is not a valid method.")

    def _get_energy_axis(self, dataset):
        """Energy axis."""
        if self.energy_edges is None:
            energy_axis = dataset.counts.geom.axes["energy"].squash()
            if getattr(self, "sum_over_energy_groups", False):
                energy_edges = [energy_axis.edges[0], energy_axis.edges[1]]
                energy_axis = MapAxis.from_energy_edges(energy_edges)
        else:
            energy_axis = MapAxis.from_energy_edges(self.energy_edges)

        return energy_axis

    def copy(self):
        """Copy estimator."""
        return deepcopy(self)

    @property
    def config_parameters(self):
        """Configuration parameters."""
        pars = self.__dict__.copy()
        pars = {key.strip("_"): value for key, value in pars.items()}
        return pars

    def __str__(self):
        s = f"{self.__class__.__name__}\n"
        s += "-" * (len(s) - 1) + "\n\n"

        pars = self.config_parameters
        max_len = np.max([len(_) for _ in pars]) + 1

        for name, value in sorted(pars.items()):
            if isinstance(value, ModelBase):
                s += f"\t{name:{max_len}s}: {value.tag[0]}\n"
            elif inspect.isclass(value):
                s += f"\t{name:{max_len}s}: {value.__name__}\n"
            elif isinstance(value, u.Quantity):
                s += f"\t{name:{max_len}s}: {value}\n"
            elif isinstance(value, Estimator):
                pass
            elif isinstance(value, np.ndarray):
                value = np.array_str(value, precision=2, suppress_small=True)
                s += f"\t{name:{max_len}s}: {value}\n"
            else:
                s += f"\t{name:{max_len}s}: {value}\n"

        return s.expandtabs(tabsize=2)

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import warnings
from astropy.io import fits
from astropy.table import Table
import astropy.units as u
from gammapy.data import EventListMetaData, EventList
from gammapy.utils.scripts import make_path
from gammapy.utils.metadata import CreatorMetaData
from gammapy.utils.time import time_ref_from_dict


class EventListReader:
    """Reader class for EventList.

    Format specification: :ref:`gadf:iact-events`

    Parameters
    ----------
    hdu : str
        Name of events HDU. Default is "EVENTS".
    checksum : bool
        If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.
    """

    def __init__(self, hdu="EVENTS", checksum=False):
        self.hdu = hdu
        self.checksum = checksum

    @staticmethod
    def from_gadf_hdu(events_hdu):
        """Create EventList from gadf HDU."""
        table = Table.read(events_hdu)
        meta = EventListMetaData.from_header(table.meta)

        # This is not a strict check on input. It just checks that required information is there.
        required_colnames = set(["RA", "DEC", "TIME", "ENERGY"])
        if not required_colnames.issubset(set(table.colnames)):
            missing_columns = required_colnames.difference(set(table.colnames))
            raise ValueError(
                f"GADF event table does not contain required columns {missing_columns}"
            )

        met = u.Quantity(table["TIME"].astype("float64"), "second")
        time = time_ref_from_dict(table.meta) + met

        energy = table["ENERGY"].quantity

        ra = table["RA"].quantity
        dec = table["DEC"].quantity

        removed_colnames = ["RA", "DEC", "GLON", "GLAT", "TIME", "ENERGY"]

        new_table = Table(
            {"TIME": time, "ENERGY": energy, "RA": ra, "DEC": dec}, meta=table.meta
        )
        for name in table.colnames:
            if name not in removed_colnames:
                new_table.add_column(table[name])

        return EventList(new_table, meta)

    @staticmethod
    def identify_format_from_hduclass(events_hdu):
        """Identify format for HDU header keywords."""
        hduclass = events_hdu.header.get("HDUCLASS", "unknown")
        return hduclass.lower()

    def read(self, filename, format="gadf"):
        """Read EventList from file.

        Parameters
        ----------
        filename : `pathlib.Path`, str
            Filename
        format : {"gadf"}, optional
            format of the EventList. Default is 'gadf'.
            If None, will try to guess from header.
        """
        filename = make_path(filename)

        with fits.open(filename) as hdulist:
            events_hdu = hdulist[self.hdu]

            if self.checksum:
                if events_hdu.verify_checksum() != 1:
                    warnings.warn(
                        f"Checksum verification failed for HDU {self.hdu} of {filename}.",
                        UserWarning,
                    )

            if format is None:
                format = self.identify_format_from_hduclass(events_hdu)

            if format == "gadf" or format == "ogip":
                return self.from_gadf_hdu(events_hdu)
            else:
                raise ValueError(f"Unknown format :{format}")


class EventListWriter:
    """Writer class for EventList."""

    def __init__(self):
        pass

    @staticmethod
    def _to_gadf_table_hdu(event_list):
        """Convert input event list to a `~astropy.io.fits.BinTableHDU` according gadf."""
        gadf_table = event_list.table.copy()
        gadf_table.remove_column("TIME")

        reference_time = time_ref_from_dict(gadf_table.meta)
        gadf_table["TIME"] = (event_list.time - reference_time).to("s")

        bin_table = fits.BinTableHDU(gadf_table, name="EVENTS")

        # A priori don't change creator information
        if event_list.meta.creation is None:
            event_list.meta.creation = CreatorMetaData()
        else:
            event_list.meta.creation.update_time()

        bin_table.header.update(event_list.meta.to_header())
        return bin_table

    def to_hdu(self, event_list, format="gadf"):
        """
        Convert input event list to a `~astropy.io.fits.BinTableHDU` according to format.

        Parameters
        ----------
        format : str, optional
            Output format, currently only "gadf" is supported. Default is "gadf".

        Returns
        -------
        hdu : `astropy.io.fits.BinTableHDU`
            EventList converted to FITS representation.
        """
        if format != "gadf":
            raise ValueError(f"Only the 'gadf' format supported, got {format}")

        return self._to_gadf_table_hdu(event_list)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""FoV background estimation."""

import logging
import numpy as np
from gammapy.maps import Map, RegionGeom
from gammapy.modeling import Fit
from gammapy.modeling.models import FoVBackgroundModel, Model
from ..core import Maker

__all__ = ["FoVBackgroundMaker"]

log = logging.getLogger(__name__)


class FoVBackgroundMaker(Maker):
    """Normalize template background on the whole field-of-view.

    The dataset background model can be simply scaled (method="scale") or fitted
    (method="fit") on the dataset counts.

    The normalization is performed outside the exclusion mask that is passed on
    init. This also internally takes into account the dataset fit mask.

    If a SkyModel is set on the input dataset its parameters
    are frozen during the FoV re-normalization.

    If the requirement (greater than) of either min_counts or min_npred_background
    is not satisfied, the background will not be normalised.

    Parameters
    ----------
    method : {'scale', 'fit'}
        The normalization method to be applied. Default 'scale'.
    exclusion_mask : `~gammapy.maps.WcsNDMap`
        Exclusion mask.
    spectral_model : `~gammapy.modeling.models.SpectralModel` or str, optional
        Reference norm spectral model to use for the `~gammapy.modeling.models.FoVBackgroundModel`, if
        none is defined on the dataset. Default is "pl-norm".
    spatial_model : `~gammapy.modeling.models.SpatialModel` or str, optional
        Spatial model to use for the `~gammapy.modeling.models.FoVBackgroundModel`, if
        none is defined on the dataset. Default is None.
        The unit of the spatial model is dropped.
    min_counts : int, optional
        Minimum number of counts, or residuals counts if a `~gammapy.modeling.models.SkyModel`
        is set, required outside the exclusion region. Default is 0.
    min_npred_background : float, optional
        Minimum number of predicted background counts required outside the
        exclusion region. Default is 0.
    """

    tag = "FoVBackgroundMaker"
    available_methods = ["fit", "scale"]

    def __init__(
        self,
        method="scale",
        exclusion_mask=None,
        spectral_model="pl-norm",
        spatial_model=None,
        min_counts=0,
        min_npred_background=0,
        fit=None,
    ):
        self.method = method
        self.exclusion_mask = exclusion_mask
        self.min_counts = min_counts
        self.min_npred_background = min_npred_background

        if isinstance(spectral_model, str):
            spectral_model = Model.create(tag=spectral_model, model_type="spectral")

        if isinstance(spatial_model, str):
            spatial_model = Model.create(tag=spatial_model, model_type="spatial")

        if not spectral_model.is_norm_spectral_model:
            raise ValueError("Spectral model must be a norm spectral model")

        self.default_spectral_model = spectral_model
        self.default_spatial_model = spatial_model

        if fit is None:
            fit = Fit()

        self.fit = fit

    @property
    def method(self):
        """Method property."""
        return self._method

    @method.setter
    def method(self, value):
        """Method setter."""
        if value not in self.available_methods:
            raise ValueError(
                f"Not a valid method for FoVBackgroundMaker: {value}."
                f" Choose from {self.available_methods}"
            )

        self._method = value

    def make_default_fov_background_model(self, dataset):
        """Add FoV background model to the model definition.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Input map dataset.

        Returns
        -------
        dataset : `~gammapy.datasets.MapDataset`
            Map dataset including background model.

        """
        if self.default_spatial_model:
            spatial_model = self.default_spatial_model.copy()
        else:
            spatial_model = None

        bkg_model = FoVBackgroundModel(
            dataset_name=dataset.name,
            spectral_model=self.default_spectral_model.copy(),
            spatial_model=spatial_model,
        )

        if dataset.models is None:
            dataset.models = bkg_model
        else:
            dataset.models = dataset.models + bkg_model

        return dataset

    def make_exclusion_mask(self, dataset):
        """Project input exclusion mask to dataset geometry.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Input map dataset.

        Returns
        -------
        mask : `~gammapy.maps.WcsNDMap`
            Projected exclusion mask.
        """
        geom = dataset._geom
        if self.exclusion_mask:
            mask = self.exclusion_mask.interp_to_geom(geom=geom)
        else:
            mask = Map.from_geom(geom=geom, data=True, dtype=bool)
        return mask

    def _make_masked_summed_counts(self, dataset):
        """Compute the sums of the counts, npred, and background maps within the mask."""
        npred = dataset.npred()
        mask = dataset.mask & ~np.isnan(npred)
        count_tot = dataset.counts.data[mask].sum()
        npred_tot = npred.data[mask].sum()
        bkg_tot = dataset.npred_background().data[mask].sum()
        return {
            "counts": count_tot,
            "npred": npred_tot,
            "bkg": bkg_tot,
        }

    def _verify_requirements(self, dataset):
        """Verify that the requirements of min_counts and min_npred_background are satisfied."""
        total = self._make_masked_summed_counts(dataset)
        not_bkg_tot = total["npred"] - total["bkg"]

        value = (total["counts"] - not_bkg_tot) / total["bkg"]
        if not np.isfinite(value):
            log.warning(
                f"FoVBackgroundMaker failed. Non-finite normalisation value for {dataset.name}. "
                "Setting mask to False."
            )
            return False
        elif total["bkg"] <= self.min_npred_background:
            log.warning(
                f"FoVBackgroundMaker failed. Only {int(total['bkg'])} background"
                f" counts outside exclusion mask for {dataset.name}. "
                "Setting mask to False."
            )
            return False
        elif total["counts"] <= self.min_counts:
            log.warning(
                f"FoVBackgroundMaker failed. Only {int(total['counts'])} counts "
                f"outside exclusion mask for {dataset.name}. "
                "Setting mask to False."
            )
            return False
        elif total["counts"] - not_bkg_tot <= 0:
            log.warning(
                f"FoVBackgroundMaker failed. Negative residuals counts for {dataset.name}. "
                "Setting mask to False."
            )
            return False
        else:
            return True

    def run(self, dataset, observation=None):
        """Run FoV background maker.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Input map dataset.

        """
        if isinstance(dataset.counts.geom, RegionGeom):
            raise TypeError(
                "FoVBackgroundMaker does not support region based datasets."
            )

        mask_fit = dataset.mask_fit
        if mask_fit:
            dataset.mask_fit *= self.make_exclusion_mask(dataset)
        else:
            dataset.mask_fit = self.make_exclusion_mask(dataset)

        if dataset.background_model is None:
            dataset = self.make_default_fov_background_model(dataset)

        if self._verify_requirements(dataset) is True:
            if self.method == "fit":
                dataset = self.make_background_fit(dataset)
            else:
                # always scale the background first
                dataset = self.make_background_scale(dataset)
        else:
            dataset.mask_safe.data[...] = False

        dataset.mask_fit = mask_fit
        return dataset

    def make_background_fit(self, dataset):
        """Fit the FoV background model on the dataset counts data.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Input dataset.

        Returns
        -------
        dataset : `~gammapy.datasets.MapDataset`
            Map dataset with fitted background model.
        """
        # freeze all model components not related to background model
        models = dataset.models.select(tag="sky-model")

        with models.restore_status(restore_values=False):
            models.select(tag="sky-model").freeze()

            fit_result = self.fit.run(datasets=[dataset])
            if not fit_result.success:
                log.warning(
                    f"FoVBackgroundMaker failed. Fit did not converge for {dataset.name}. "
                    "Setting mask to False."
                )
                dataset.mask_safe.data[...] = False

        return dataset

    def make_background_scale(self, dataset):
        """Fit the FoV background model on the dataset counts data.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Input dataset.

        Returns
        -------
        dataset : `~gammapy.datasets.MapDataset`
            Map dataset with scaled background model.
        """
        total = self._make_masked_summed_counts(dataset)
        not_bkg_tot = total["npred"] - total["bkg"]

        value = (total["counts"] - not_bkg_tot) / total["bkg"]
        error = np.sqrt(total["counts"] - not_bkg_tot) / total["bkg"]
        dataset.models[f"{dataset.name}-bkg"].spectral_model.norm.value = value
        dataset.models[f"{dataset.name}-bkg"].spectral_model.norm.error = error

        return dataset



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.time import Time
from gammapy.modeling.models import LightCurveTemplateTemporalModel
from gammapy.modeling.models.utils import _template_model_from_cta_sdc, read_hermes_cube
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import requires_data, requires_dependency


@requires_data()
def test__template_model_from_cta_sdc(tmp_path):
    filename = "$GAMMAPY_DATA/gravitational_waves/GW_example_DC_file.fits.gz"
    mod = _template_model_from_cta_sdc(filename)

    assert isinstance(mod, LightCurveTemplateTemporalModel)
    t = Time(55555.6157407407, format="mjd")
    val = mod.evaluate(t, energy=[0.3, 2] * u.TeV)
    assert_allclose(val.data, [[2.281216e-21], [4.281390e-23]], rtol=1e-5)

    filename = make_path(tmp_path / "test.fits")
    mod.write(filename=filename, format="map", overwrite=True)
    mod1 = LightCurveTemplateTemporalModel.read(filename=filename, format="map")

    assert_allclose(mod1.t_ref.value, mod.t_ref.value, rtol=1e-7)


@requires_data()
def test__reference_time():
    filename = "$GAMMAPY_DATA/gravitational_waves/GW_example_DC_file.fits.gz"
    t_ref = Time("2028-01-01T00:00:00", format="isot", scale="utc")
    mod = _template_model_from_cta_sdc(filename, t_ref=t_ref)

    assert_allclose(mod.reference_time.mjd, t_ref.mjd, rtol=1e-7)


@requires_data()
@requires_dependency("healpy")
def test_read_hermes_cube():
    filename = make_path(
        "$GAMMAPY_DATA/tests/hermes/hermes-VariableMin-pi0-Htot_CMZ_nside256.fits.gz"
    )
    map_ = read_hermes_cube(filename)
    assert map_.geom.frame == "galactic"
    assert map_.geom.axes[0].unit == "GeV"
    assert_allclose(map_.geom.axes[0].center[3], 1 * u.GeV)
    assert_allclose(map_.get_by_coord((0 * u.deg, 0 * u.deg, 1 * u.GeV)), 2.6391575)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import astropy.units as u
from .core import PSF

__all__ = ["PSF3D"]

log = logging.getLogger(__name__)


class PSF3D(PSF):
    """PSF with axes: energy, offset, rad.

    Data format specification: :ref:`gadf:psf_table`

    Parameters
    ----------
    axes : list of `~gammapy.maps.MapAxis` or `~gammapy.maps.MapAxes`
        Required axes (in the given order) are:
            * energy_true (true energy axis)
            * migra (energy migration axis)
            * rad (rad axis)
    data : `~astropy.units.Quantity`
        PSF (3-dim with axes: psf[rad_index, offset_index, energy_index].
    unit : dict of str or `~astropy.units.Unit`
        Unit.
    is_pointlike : bool, optional
        Whether the IRF is point-like.
        True for point-like IRFs, False for full-containment.
        Default is False.
    fov_alignment : `~gammapy.irf.FoVAlignment`, optional
        The orientation of the field of view coordinate system.
        Default is FoVAlignment.RADEC.
    meta : dict
        Metadata dictionary.
    """

    tag = "psf_table"
    required_axes = ["energy_true", "offset", "rad"]
    default_unit = u.sr**-1



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.io import fits
from astropy.units import Quantity
from gammapy.irf import (
    Background3D,
    EffectiveAreaTable2D,
    EnergyDependentMultiGaussPSF,
    EnergyDispersion2D,
    RadMax2D,
    load_irf_dict_from_file,
)
from gammapy.maps import MapAxis
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import requires_data


@requires_data()
def test_load_irf_dict_from_file():
    """Test that the IRF components in a dictionary loaded from a DL3 file can
    be loaded in a dictionary and correctly used"""
    irf = load_irf_dict_from_file(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
    )

    energy = Quantity(1, "TeV")
    offset = Quantity(0.5, "deg")

    val = irf["aeff"].evaluate(energy_true=energy, offset=offset)
    assert_allclose(val.value, 273372.44851054, rtol=1e-5)
    assert val.unit == "m2"

    val = irf["edisp"].evaluate(offset=offset, energy_true=energy, migra=1)
    assert_allclose(val.value, 1.84269482, rtol=1e-5)
    assert val.unit == ""

    val = irf["psf"].evaluate(
        rad=Quantity(0.1, "deg"), energy_true=energy, offset=offset
    )
    assert_allclose(val, 6.75981573 * u.Unit("deg-2"), rtol=1e-5)

    val = irf["bkg"].evaluate(energy=energy, fov_lon=offset, fov_lat="0.1 deg")
    assert_allclose(val.value, 0.00031552, rtol=1e-5)
    assert val.unit == "1 / (MeV s sr)"


@requires_data()
def test_irf_dict_from_file_duplicate_irfs(caplog, tmp_path):
    """catch the warning message about two type of IRF with the same hdu class
    encountered in the same file"""
    original_file = make_path(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
    )
    dummy_file = tmp_path / "020136_duplicated_psf.fits"

    # create a dummy file with the PSF HDU repeated twice
    f = fits.open(original_file)
    f.append(f[5].copy())
    f[7].name = "PSF2"
    f.writeto(dummy_file)

    load_irf_dict_from_file(dummy_file)

    assert "more than one HDU" in caplog.text
    assert "loaded the PSF HDU in the dictionary" in caplog.text


@requires_data()
def test_irf_dict_from_file_fixed_rad_max():
    """test that for point-like IRF without RAD_MAX_2D HDU a RadMax2D with a
    single value is generated from the RAD_MAX header keyword"""
    irf = load_irf_dict_from_file(
        "$GAMMAPY_DATA/joint-crab/dl3/magic/run_05029748_DL3.fits"
    )

    assert "RAD_MAX" in irf["aeff"].meta
    assert "rad_max" in irf
    assert isinstance(irf["rad_max"], RadMax2D)

    # check that has a single-bin in energy and offset
    assert irf["rad_max"].axes["energy"].nbin == 1
    assert irf["rad_max"].axes["offset"].nbin == 1
    assert irf["rad_max"].quantity.to_value("deg") == irf["aeff"].meta["RAD_MAX"]


class TestIRFWrite:
    def setup_method(self):
        self.energy_lo = np.logspace(0, 1, 10)[:-1] * u.TeV
        self.energy_hi = np.logspace(0, 1, 10)[1:] * u.TeV
        self.energy_axis_true = MapAxis.from_energy_bounds(
            "1 TeV", "10 TeV", nbin=9, name="energy_true"
        )

        self.offset_lo = np.linspace(0, 1, 4)[:-1] * u.deg
        self.offset_hi = np.linspace(0, 1, 4)[1:] * u.deg

        self.offset_axis = MapAxis.from_bounds(
            0, 1, nbin=3, unit="deg", name="offset", node_type="edges"
        )
        self.migra_lo = np.linspace(0, 3, 4)[:-1]
        self.migra_hi = np.linspace(0, 3, 4)[1:]
        self.migra_axis = MapAxis.from_bounds(
            0, 3, nbin=3, name="migra", node_type="edges"
        )
        self.fov_lon_lo = np.linspace(-6, 6, 11)[:-1] * u.deg
        self.fov_lon_hi = np.linspace(-6, 6, 11)[1:] * u.deg
        self.fov_lon_axis = MapAxis.from_bounds(-6, 6, nbin=10, name="fov_lon")

        self.fov_lat_lo = np.linspace(-6, 6, 11)[:-1] * u.deg
        self.fov_lat_hi = np.linspace(-6, 6, 11)[1:] * u.deg
        self.fov_lat_axis = MapAxis.from_bounds(-6, 6, nbin=10, name="fov_lat")

        self.aeff_data = np.random.rand(9, 3) * u.cm * u.cm
        self.edisp_data = np.random.rand(9, 3, 3)
        self.bkg_data = np.random.rand(9, 10, 10) / u.MeV / u.s / u.sr

        self.aeff = EffectiveAreaTable2D(
            axes=[self.energy_axis_true, self.offset_axis],
            data=self.aeff_data.value,
            unit=self.aeff_data.unit,
        )
        self.edisp = EnergyDispersion2D(
            axes=[
                self.energy_axis_true,
                self.migra_axis,
                self.offset_axis,
            ],
            data=self.edisp_data,
        )
        axes = [
            self.energy_axis_true.copy(name="energy"),
            self.fov_lon_axis,
            self.fov_lat_axis,
        ]
        self.bkg = Background3D(
            axes=axes, data=self.bkg_data.value, unit=self.bkg_data.unit
        )

    def test_array_to_container(self):
        assert_allclose(self.aeff.quantity, self.aeff_data)
        assert_allclose(self.edisp.quantity, self.edisp_data)
        assert_allclose(self.bkg.quantity, self.bkg_data)

    def test_container_to_table(self):
        assert_allclose(self.aeff.to_table()["ENERG_LO"].quantity[0], self.energy_lo)
        assert_allclose(self.edisp.to_table()["ENERG_LO"].quantity[0], self.energy_lo)
        assert_allclose(self.bkg.to_table()["ENERG_LO"].quantity[0], self.energy_lo)

        assert_allclose(self.aeff.to_table()["EFFAREA"].quantity[0].T, self.aeff_data)
        assert_allclose(self.edisp.to_table()["MATRIX"].quantity[0].T, self.edisp_data)
        assert_allclose(self.bkg.to_table()["BKG"].quantity[0].T, self.bkg_data)

        assert self.aeff.to_table()["EFFAREA"].quantity[0].unit == self.aeff_data.unit
        assert self.bkg.to_table()["BKG"].quantity[0].unit == self.bkg_data.unit

    def test_container_to_fits(self):
        assert_allclose(self.aeff.to_table()["ENERG_LO"].quantity[0], self.energy_lo)

        assert self.aeff.to_table_hdu().header["EXTNAME"] == "EFFECTIVE AREA"
        assert self.edisp.to_table_hdu().header["EXTNAME"] == "ENERGY DISPERSION"
        assert self.bkg.to_table_hdu().header["EXTNAME"] == "BACKGROUND"

        hdu = self.aeff.to_table_hdu()
        assert_allclose(
            hdu.data[hdu.header["TTYPE1"]][0], self.aeff.axes[0].edges[:-1].value
        )
        hdu = self.aeff.to_table_hdu()
        assert_allclose(hdu.data[hdu.header["TTYPE5"]][0].T, self.aeff.data)

        hdu = self.edisp.to_table_hdu()
        assert_allclose(
            hdu.data[hdu.header["TTYPE1"]][0], self.edisp.axes[0].edges[:-1].value
        )
        hdu = self.edisp.to_table_hdu()
        assert_allclose(hdu.data[hdu.header["TTYPE7"]][0].T, self.edisp.data)

        hdu = self.bkg.to_table_hdu()
        assert_allclose(
            hdu.data[hdu.header["TTYPE1"]][0], self.bkg.axes[0].edges[:-1].value
        )
        hdu = self.bkg.to_table_hdu()
        assert_allclose(hdu.data[hdu.header["TTYPE7"]][0].T, self.bkg.data)

    def test_writeread(self, tmp_path):
        path = tmp_path / "tmp.fits"
        fits.HDUList(
            [
                fits.PrimaryHDU(),
                self.aeff.to_table_hdu(),
                self.edisp.to_table_hdu(),
                self.bkg.to_table_hdu(),
            ]
        ).writeto(path)

        read_aeff = EffectiveAreaTable2D.read(path, hdu="EFFECTIVE AREA")
        assert_allclose(read_aeff.quantity, self.aeff_data)

        read_edisp = EnergyDispersion2D.read(path, hdu="ENERGY DISPERSION")
        assert_allclose(read_edisp.quantity, self.edisp_data)

        read_bkg = Background3D.read(path, hdu="BACKGROUND")
        assert_allclose(read_bkg.quantity, self.bkg_data)


@requires_data()
def test_load_irf_dict_from_file_cta():
    """Test that CTA IRFs can be loaded and evaluated."""
    irf = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    assert set(irf.keys()) == {"aeff", "edisp", "psf", "bkg"}
    assert isinstance(irf["aeff"], EffectiveAreaTable2D)
    assert isinstance(irf["edisp"], EnergyDispersion2D)
    assert isinstance(irf["psf"], EnergyDependentMultiGaussPSF)
    assert isinstance(irf["bkg"], Background3D)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import importlib
from enum import Enum
import logging

log = logging.getLogger(__name__)


def is_numba_available():
    """Check if ray is available."""
    try:
        importlib.import_module("numba")
        return True
    except ModuleNotFoundError:
        return False


class CompilationBackendEnum(Enum):
    """Enum for parallel backend."""

    cython = "cython"
    jit = "jit"

    @classmethod
    def from_str(cls, value):
        """Get enum from string."""
        if value == "jit" and not is_numba_available():
            log.warning("numba is not installed, falling back to cython backend")
            value = "cython"

        return cls(value)


def _get_fit_statistics_cython():
    """Get fit_statistics module with cython."""
    from gammapy.stats.fit_statistics_cython import (
        TRUNCATION_VALUE,
        weighted_cash_sum_cython,
        cash_sum_cython,
        f_cash_root_cython,
        norm_bounds_cython,
    )

    return dict(
        TRUNCATION_VALUE=TRUNCATION_VALUE,
        weighted_cash_sum_compiled=weighted_cash_sum_cython,
        cash_sum_compiled=cash_sum_cython,
        f_cash_root_compiled=f_cash_root_cython,
        norm_bounds_compiled=norm_bounds_cython,
    )


def _get_fit_statistics_jit():
    """Get fit_statistics module with numba backend."""
    from gammapy.stats.fit_statistics_jit import (
        TRUNCATION_VALUE,
        weighted_cash_sum_jit,
        cash_sum_jit,
        f_cash_root_jit,
        norm_bounds_jit,
    )

    return dict(
        TRUNCATION_VALUE=TRUNCATION_VALUE,
        weighted_cash_sum_compiled=weighted_cash_sum_jit,
        cash_sum_compiled=cash_sum_jit,
        f_cash_root_compiled=f_cash_root_jit,
        norm_bounds_compiled=norm_bounds_jit,
    )


COMPILATION_BACKEND_DEFAULT = CompilationBackendEnum.cython

COMPILED_STATS_MODULES = {
    CompilationBackendEnum.cython: _get_fit_statistics_cython,
    CompilationBackendEnum.jit: _get_fit_statistics_jit,
}


def get_fit_statistics_compiled(backend=None):
    if backend is None:
        from gammapy.utils.compilation import COMPILATION_BACKEND_DEFAULT

        backend = COMPILATION_BACKEND_DEFAULT
    backend = CompilationBackendEnum.from_str(backend)
    return COMPILED_STATS_MODULES[backend]()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import html
import logging
import warnings
from enum import Enum, auto
import numpy as np
import scipy.interpolate
import astropy.units as u
from astropy.coordinates import (
    ICRS,
    AltAz,
    BaseCoordinateFrame,
    CartesianRepresentation,
    SkyCoord,
    UnitSphericalRepresentation,
)
from astropy.io import fits
from astropy.table import Table
from astropy.units import Quantity
from astropy.utils import lazyproperty
from gammapy.utils.compat import COPY_IF_NEEDED
from gammapy.utils.deprecation import GammapyDeprecationWarning
from gammapy.utils.fits import earth_location_from_dict
from gammapy.utils.scripts import make_path
from gammapy.utils.time import time_ref_from_dict

log = logging.getLogger(__name__)

__all__ = ["FixedPointingInfo", "PointingInfo", "PointingMode"]


def _check_coord_frame(coord_or_frame, expected_frame, name):
    """Check if a skycoord or frame is given in expected_frame."""
    is_coord = isinstance(coord_or_frame, SkyCoord)
    is_frame = isinstance(coord_or_frame, BaseCoordinateFrame)

    if not (is_frame or is_coord):
        raise TypeError(
            f"{name} must be a 'astropy.coordinates.SkyCoord'"
            "or {expected_frame} instance"
        )

    if is_coord:
        frame = coord_or_frame.frame
    else:
        frame = coord_or_frame

    if not isinstance(frame, expected_frame):
        raise ValueError(
            f"{name} is in wrong frame, expected {expected_frame}, got {frame}"
        )


class PointingMode(Enum):
    """
    Describes the behavior of the pointing during the observation.

    See :ref:`gadf:iact-events-obs-mode`.

    For ground-based instruments, the most common options will be:
    * POINTING: The telescope observes a fixed position in the ICRS frame
    * DRIFT: The telescope observes a fixed position in the alt-az frame

    Gammapy only supports fixed pointing positions over the whole observation
    (either in equatorial or horizontal coordinates).
    OGIP also defines RASTER, SLEW and SCAN. These cannot be treated using
    a fixed pointing position in either frame, so they would require the
    pointing table, which is at the moment not supported by gammapy.

    Data releases based on gadf v0.2 do not have consistent OBS_MODE keyword
    e.g. the H.E.S.S. data releases uses the not-defined value "WOBBLE".
    For all gadf data, we assume OBS_MODE to be the same as "POINTING",
    unless it is set to "DRIFT", making the assumption that one observation
    only contains a single fixed position.
    """

    POINTING = auto()
    DRIFT = auto()

    @staticmethod
    def from_gadf_string(val):
        """Parse a string from the GADF header into a PointingMode."""
        # OBS_MODE is not well-defined and not mandatory in GADF 0.2
        # We always assume that the observations are pointing observations
        # unless the OBS_MODE is set to DRIFT
        if val.upper() == "DRIFT":
            return PointingMode.DRIFT
        else:
            return PointingMode.POINTING


class FixedPointingInfo:
    """IACT array pointing info.

    Data format specification: :ref:`gadf:iact-pnt`

    Parameters
    ----------
    meta : `~astropy.table.Table.meta`
        Meta header info from Table on pointing.
        Passing this is deprecated, provide ``mode`` and ``fixed_icrs`` or ``fixed_altaz``
        instead or use `FixedPointingInfo.from_fits_header` instead.
    mode : `PointingMode`
        How the telescope was pointing during the observation.
    fixed_icrs : `~astropy.coordinates.SkyCoord`, optional
        The coordinates of the observation in ICRS as a `~astropy.coordinates.SkyCoord` object. Default is None.
        Required if mode is `PointingMode.POINTING`.
    fixed_altaz : `~astropy.coordinates.SkyCoord`, optional
        The coordinates of the observation in alt-az as a `~astropy.coordinates.SkyCoord` object. Default is None.
        Required if mode is `PointingMode.DRIFT`.

    Examples
    --------
    >>> from gammapy.data import FixedPointingInfo, PointingMode
    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u
    >>> fixed_icrs = SkyCoord(83.633 * u.deg, 22.014 * u.deg, frame="icrs")
    >>> pointing_info = FixedPointingInfo(fixed_icrs=fixed_icrs)
    >>> print(pointing_info)
    FixedPointingInfo:
    <BLANKLINE>
    mode:        PointingMode.POINTING
    coordinates: <SkyCoord (ICRS): (ra, dec) in deg
        (83.633, 22.014)>
    """

    def __init__(self, fixed_icrs=None, fixed_altaz=None):
        if fixed_icrs is not None and fixed_altaz is not None:
            raise ValueError("fixed_icrs and fixed_altaz are mutually exclusive")

        if fixed_icrs is not None:
            _check_coord_frame(fixed_icrs, ICRS, "fixed_icrs")

            if np.isnan(fixed_icrs.ra.value) or np.isnan(fixed_icrs.dec.value):
                warnings.warn(
                    "In future, fixed_icrs must have non-nan values",
                    GammapyDeprecationWarning,
                )

            self._mode = PointingMode.POINTING
            self._fixed_icrs = fixed_icrs
            self._fixed_altaz = None

        else:
            _check_coord_frame(fixed_altaz, AltAz, "fixed_altaz")
            self._mode = PointingMode.DRIFT
            self._fixed_icrs = None
            self._fixed_altaz = fixed_altaz

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    @classmethod
    def from_fits_header(cls, header):
        """
        Parse `~gammapy.data.FixedPointingInfo` from the given FITS header.

        Parameters
        ----------
        header : `astropy.fits.Header`
            Header to parse, e.g. from a GADF EVENTS HDU.
            Currently, only the GADF format is supported.

        Returns
        -------
        pointing : `~gammapy.data.FixedPointingInfo`
            The FixedPointingInfo instance filled from the given header.
        """
        obs_mode = header.get("OBS_MODE", "POINTING")
        mode = PointingMode.from_gadf_string(obs_mode)

        fixed_icrs = None
        pointing_altaz = None

        # we can be more strict with DRIFT, as support was only added recently
        if mode is PointingMode.DRIFT:
            alt = header.get("ALT_PNT", None)
            az = header.get("AZ_PNT", None)
            if alt is None or az is None:
                raise IOError(
                    "Keywords ALT_PNT and AZ_PNT are required for OBSMODE=DRIFT"
                )
            pointing_altaz = AltAz(alt=alt * u.deg, az=az * u.deg)
        else:
            ra = u.Quantity(header.get("RA_PNT", np.nan), u.deg)
            dec = u.Quantity(header.get("DEC_PNT", np.nan), u.deg)

            fixed_icrs = SkyCoord(ra, dec)
            if np.isnan(ra.value) or np.isnan(dec.value):
                raise IOError(
                    "Keywords RA_PNT / DEC_PNT are required for pointing-mode POINTING"
                )

        return cls(
            fixed_icrs=fixed_icrs,
            fixed_altaz=pointing_altaz,
        )

    def to_fits_header(self, format="gadf", version="0.3", time_ref=None):
        """
        Convert this FixedPointingInfo object into a fits header for the given format.

        Parameters
        ----------
        format : {"gadf"}
            Format, currently only "gadf" is supported. Default is "gadf".
        version : str, optional
            Version of the ``format``, this function currently supports
            gadf versions 0.2 and 0.3. Default is "0.3".
        time_ref : `astropy.time.Time`, optional
            Reference time for storing the time related information in fits format. Default is None.

        Returns
        -------
        header : `astropy.fits.Header`
            Header with fixed pointing information filled for the requested format.
        """
        if format != "gadf":
            raise ValueError(f'Only the "gadf" format supported, got {format}')

        if version not in {"0.2", "0.3"}:
            raise ValueError(f"Unsupported version {version} for format {format}")

        if self.mode == PointingMode.DRIFT and version == "0.2":
            raise ValueError("mode=DRIFT is only supported by GADF 0.3")

        header = fits.Header()
        if self.mode is PointingMode.POINTING:
            header["OBS_MODE"] = "POINTING"
            header["RA_PNT"] = self.fixed_icrs.ra.deg, u.deg.to_string("fits")
            header["DEC_PNT"] = self.fixed_icrs.dec.deg, u.deg.to_string("fits")
        elif self.mode is PointingMode.DRIFT:
            header["OBS_MODE"] = "DRIFT"
            header["AZ_PNT"] = self.fixed_altaz.az.deg, u.deg.to_string("fits")
            header["ALT_PNT"] = self.fixed_altaz.alt.deg, u.deg.to_string("fits")

        return header

    @classmethod
    def read(cls, filename, hdu="EVENTS"):
        """Read pointing information table from file to obtain the metadata.

        Parameters
        ----------
        filename : str
            Filename.
        hdu : int or str, optional
            HDU number or name. Default is "EVENTS".

        Returns
        -------
        pointing_info : `PointingInfo`
            Pointing information.
        """
        filename = make_path(filename)
        header = fits.getheader(filename, extname=hdu)
        return cls.from_fits_header(header)

    @property
    def mode(self):
        """See `PointingMode`, if not present, assume POINTING."""
        return self._mode

    @property
    def fixed_altaz(self):
        """The fixed coordinates of the observation in alt-az as a `~astropy.coordinates.SkyCoord` object.

        None if not a DRIFT observation.
        """
        return self._fixed_altaz

    @property
    def fixed_icrs(self):
        """
        The fixed coordinates of the observation in ICRS as a `~astropy.coordinates.SkyCoord` object.

        None if not a POINTING observation.
        """
        return self._fixed_icrs

    def get_icrs(self, obstime=None, location=None) -> SkyCoord:
        """
        Get the pointing position in ICRS frame for a given time.

        If the observation was performed tracking a fixed position in ICRS,
        the icrs pointing is returned with the given obstime attached.

        If the observation was performed in drift mode, the fixed alt-az coordinates
        are transformed to ICRS using the observation location and the given time.


        Parameters
        ----------
        obstime : `astropy.time.Time`, optional
            Time for which to get the pointing position in ICRS frame. Default is None.
        location : `astropy.coordinates.EarthLocation`, optional
            Observatory location, only needed for drift observations to transform
            from horizontal coordinates to ICRS. Default is None.

        Returns
        -------
        icrs : `astropy.coordinates.SkyCoord`
            Pointing position in ICRS frame.
        """
        if self.mode == PointingMode.POINTING:
            return SkyCoord(self._fixed_icrs.data, location=location, obstime=obstime)

        if self.mode == PointingMode.DRIFT:
            if obstime is None or location is None:
                raise ValueError(
                    "FixedPointingInfo.get_icrs() in DRIFT mode requires both obstime and location inputs."
                )
            return self.get_altaz(obstime, location=location).icrs

        raise ValueError(f"Unsupported pointing mode: {self.mode}.")

    def get_altaz(self, obstime=None, location=None) -> SkyCoord:
        """
        Get the pointing position in alt-az frame for a given time.

        If the observation was performed tracking a fixed position in ICRS,
        the icrs pointing is transformed at the given time using the location
        of the observation.

        If the observation was performed in drift mode,
        the fixed alt-az coordinate is returned with `obstime` attached.

        Parameters
        ----------
        obstime : `astropy.time.Time`, optional
            Time for which to get the pointing position in alt-az frame. Default is None.
        location : `astropy.coordinates.EarthLocation`, optional
            Observatory location, only needed for pointing observations to transform
            from ICRS to horizontal coordinates. Default is None.

        Returns
        -------
        altaz : `astropy.coordinates.SkyCoord`
            Pointing position in alt-az frame.
        """
        if obstime is None or location is None:
            raise ValueError(
                "FixedPointingInfo.get_altaz() requires both obstime and location inputs."
            )

        frame = AltAz(location=location, obstime=obstime)

        if self.mode == PointingMode.POINTING:
            return self.fixed_icrs.transform_to(frame)

        if self.mode == PointingMode.DRIFT:
            # see https://github.com/astropy/astropy/issues/12965
            # TODO: should be solved in astropy v6.0. Simplify once the astropy dependency imposes >=6.0
            alt = self.fixed_altaz.alt
            az = self.fixed_altaz.az
            return SkyCoord(
                alt=u.Quantity(
                    np.full(obstime.shape, alt.deg), u.deg, copy=COPY_IF_NEEDED
                ),
                az=u.Quantity(
                    np.full(obstime.shape, az.deg), u.deg, copy=COPY_IF_NEEDED
                ),
                frame=frame,
            )

        raise ValueError(f"Unsupported pointing mode: {self.mode}.")

    def __str__(self):
        coordinates = (
            self.fixed_icrs if self.mode == PointingMode.POINTING else self.fixed_altaz
        )
        return (
            "FixedPointingInfo:\n\n"
            f"mode:        {self.mode}\n"
            f"coordinates: {coordinates}"
        )


class PointingInfo:
    """IACT array pointing info.

    Data format specification: :ref:`gadf:iact-pnt`.

    Parameters
    ----------
    table : `~astropy.table.Table`
        Table (with meta header information) on pointing.

    Examples
    --------
    >>> from gammapy.data import PointingInfo
    >>> pointing_info = PointingInfo.read('$GAMMAPY_DATA/tests/pointing_table.fits.gz')
    >>> print(pointing_info)
    Pointing info:
    <BLANKLINE>
    Location:     GeodeticLocation(lon=<Longitude 16.50022222 deg>, lat=<Latitude -23.27177778 deg>, height=<Quantity 1835. m>)
    MJDREFI, MJDREFF, TIMESYS = (51910, 0.000742870370370241, 'TT')
    Time ref:     2001-01-01T00:01:04.184
    Time ref:     51910.00074287037 MJD (TT)
    Duration:     1586.0000000000018 sec = 0.44055555555555603 hours
    Table length: 100
    <BLANKLINE>
    START:
    Time:  2004-01-21T19:50:02.184
    Time:  53025.826414166666 MJD (TT)
    RADEC: 83.6333 24.5144 deg
    ALTAZ: 11.4575 41.3409 deg
    <BLANKLINE>
    <BLANKLINE>
    END:
    Time:  2004-01-21T20:16:28.184
    Time:  53025.844770648146 MJD (TT)
    RADEC: 83.6333 24.5144 deg
    ALTAZ: 3.44573 42.1319 deg
    <BLANKLINE>
    <BLANKLINE>

    Note: In order to reproduce the example you need the tests datasets folder.
    You may download it with the command
    ``gammapy download datasets --tests --out $GAMMAPY_DATA``
    """

    def __init__(self, table):
        self.table = table

    @classmethod
    def read(cls, filename, hdu="POINTING"):
        """Read `PointingInfo` table from file.

        Parameters
        ----------
        filename : str
            Filename.
        hdu : int or str, optional
            HDU number or name. Default is "POINTING".

        Returns
        -------
        pointing_info : `PointingInfo`
            Pointing information.
        """
        filename = make_path(filename)
        table = Table.read(filename, hdu=hdu)
        return cls(table=table)

    def __str__(self):
        ss = "Pointing info:\n\n"
        ss += f"Location:     {self.location.geodetic}\n"
        m = self.table.meta
        ss += "MJDREFI, MJDREFF, TIMESYS = {}\n".format(
            (m["MJDREFI"], m["MJDREFF"], m["TIMESYS"])
        )
        ss += f"Time ref:     {self.time_ref.fits}\n"
        ss += f"Time ref:     {self.time_ref.mjd} MJD (TT)\n"
        sec = self.duration.to("second").value
        hour = self.duration.to("hour").value
        ss += f"Duration:     {sec} sec = {hour} hours\n"
        ss += "Table length: {}\n".format(len(self.table))

        ss += "\nSTART:\n" + self._str_for_index(0) + "\n"
        ss += "\nEND:\n" + self._str_for_index(-1) + "\n"

        return ss

    def _str_for_index(self, idx):
        """Information for one point in the pointing table."""
        ss = "Time:  {}\n".format(self.time[idx].fits)
        ss += "Time:  {} MJD (TT)\n".format(self.time[idx].mjd)
        ss += "RADEC: {} deg\n".format(self.radec[idx].to_string())
        ss += "ALTAZ: {} deg\n".format(self.altaz[idx].to_string())
        return ss

    @lazyproperty
    def location(self):
        """Observatory location as an `~astropy.coordinates.EarthLocation` object."""
        return earth_location_from_dict(self.table.meta)

    @lazyproperty
    def time_ref(self):
        """Time reference as a `~astropy.time.Time` object."""
        return time_ref_from_dict(self.table.meta)

    @lazyproperty
    def duration(self):
        """Pointing table duration as a `~astropy.time.TimeDelta` object.

        The time difference between the first and last entry.
        """
        return self.time[-1] - self.time[0]

    @lazyproperty
    def time(self):
        """Time array as a `~astropy.time.Time` object."""
        met = Quantity(self.table["TIME"].astype("float64"), "second")
        time = self.time_ref + met
        return time.tt

    @lazyproperty
    def radec(self):
        """RA / DEC position from table as a `~astropy.coordinates.SkyCoord`."""
        lon = self.table["RA_PNT"]
        lat = self.table["DEC_PNT"]
        return SkyCoord(lon, lat, unit="deg", frame="icrs")

    @lazyproperty
    def altaz_frame(self):
        """ALT / AZ frame as a `~astropy.coordinates.AltAz` object."""
        return AltAz(obstime=self.time, location=self.location)

    @lazyproperty
    def altaz(self):
        """ALT / AZ position computed from RA / DEC as a`~astropy.coordinates.SkyCoord`."""
        return self.radec.transform_to(self.altaz_frame)

    @lazyproperty
    def altaz_from_table(self):
        """ALT / AZ position from table as a `~astropy.coordinates.SkyCoord`."""
        lon = self.table["AZ_PNT"]
        lat = self.table["ALT_PNT"]
        return SkyCoord(lon, lat, unit="deg", frame=self.altaz_frame)

    @staticmethod
    def _interpolate_cartesian(mjd_support, coord_support, mjd):
        xyz = coord_support.cartesian
        x_new = scipy.interpolate.interp1d(mjd_support, xyz.x)(mjd)
        y_new = scipy.interpolate.interp1d(mjd_support, xyz.y)(mjd)
        z_new = scipy.interpolate.interp1d(mjd_support, xyz.z)(mjd)
        return CartesianRepresentation(x_new, y_new, z_new).represent_as(
            UnitSphericalRepresentation
        )

    def altaz_interpolate(self, time):
        """Interpolate pointing for a given time."""
        altaz_frame = AltAz(obstime=time, location=self.location)
        return SkyCoord(
            self._interpolate_cartesian(self.time.mjd, self.altaz, time.mjd),
            frame=altaz_frame,
        )

    def get_icrs(self, obstime):
        """
        Get the pointing position in ICRS frame for a given time.

        Parameters
        ----------
        obstime : `astropy.time.Time`
            Time for which to get the pointing position in ICRS frame.

        Returns
        -------
        icrs : `astropy.coordinates.SkyCoord`
            Pointing position in ICRS frame.
        """
        return SkyCoord(
            self._interpolate_cartesian(self.time.mjd, self.radec, obstime.mjd),
            obstime=obstime,
            frame="icrs",
        )

    def get_altaz(self, obstime):
        """
        Get the pointing position in alt-az frame for a given time.

        If the observation was performed tracking a fixed position in ICRS,
        the icrs pointing is transformed at the given time using the location
        of the observation.

        If the observation was performed in drift mode,
        the fixed alt-az coordinate is returned with `obstime` attached.

        Parameters
        ----------
        obstime : `astropy.time.Time`
            Time for which to get the pointing position in alt-az frame.

        Returns
        -------
        altaz : `astropy.coordinates.SkyCoord`
            Pointing position in alt-az frame.
        """
        # give precedence to ALT_PNT / AZ_PNT if present
        if "ALT_PNT" in self.table and "AZ_PNT" in self.table:
            altaz = self.altaz_from_table
            frame = AltAz(obstime=obstime, location=self.location)
            return SkyCoord(
                self._interpolate_cartesian(self.time.mjd, altaz, obstime.mjd),
                frame=frame,
            )

        # fallback to transformation from required ICRS if not
        return self.altaz_interpolate(time=obstime)



# Licensed under a 3-clause BSD style license - see LICENSE.rst


def _parse_datasets(datasets):
    """Parser used by Fit and Sampler classes."""
    from gammapy.datasets import Dataset, Datasets

    if isinstance(datasets, (list, Dataset)):
        datasets = Datasets(datasets)
    return datasets, datasets.parameters



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose, assert_equal
import astropy.units as u
from gammapy.irf import EffectiveAreaTable2D
from gammapy.maps import MapAxis
from gammapy.utils.testing import (
    assert_quantity_allclose,
    mpl_plot_check,
    requires_data,
)


@pytest.fixture(scope="session")
def aeff():
    filename = "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_023523.fits.gz"
    return EffectiveAreaTable2D.read(filename, hdu="AEFF")


@requires_data()
def test_basic(aeff):
    assert aeff.axes["energy_true"].nbin == 96
    assert aeff.axes["offset"].nbin == 6
    assert aeff.data.shape == (96, 6)

    assert aeff.axes["energy_true"].unit == "TeV"
    assert aeff.axes["offset"].unit == "deg"
    assert aeff.unit == "m2"

    assert_quantity_allclose(aeff.meta["HI_THRES"], 100, rtol=1e-3)
    assert_quantity_allclose(aeff.meta["LO_THRES"], 0.870964, rtol=1e-3)

    test_val = aeff.evaluate(energy_true="14 TeV", offset="0.2 deg")
    assert_allclose(test_val.value, 683177.5, rtol=1e-3)


def test_from_parametrization():
    # Log center of this is 100 GeV
    area_ref = 1.65469579e07 * u.cm**2

    axis = MapAxis.from_energy_edges([80, 125] * u.GeV, name="energy_true")
    area = EffectiveAreaTable2D.from_parametrization(axis, "HESS")

    assert_allclose(area.quantity, area_ref)
    assert area.unit == area_ref.unit

    # Log center of this is 0.1, 2 TeV
    area_ref = [1.65469579e07, 1.46451957e09] * u.cm * u.cm

    axis = MapAxis.from_energy_edges([0.08, 0.125, 32] * u.TeV, name="energy_true")
    area = EffectiveAreaTable2D.from_parametrization(axis, "HESS")
    assert_allclose(area.quantity[:, 0], area_ref)
    assert area.unit == area_ref.unit
    assert area.meta["TELESCOP"] == "HESS"

    area2 = EffectiveAreaTable2D.from_parametrization(axis, "CTAO")
    assert area2.unit == area_ref.unit

    with pytest.raises(ValueError):
        area2 = EffectiveAreaTable2D.from_parametrization(axis, "SWIFT")


@requires_data()
def test_plot(aeff):
    with mpl_plot_check():
        aeff.plot()

    with mpl_plot_check():
        aeff.plot_energy_dependence()

    with mpl_plot_check():
        aeff.plot_offset_dependence()


def test_to_table():
    energy_axis_true = MapAxis.from_energy_bounds(
        "1 TeV", "10 TeV", nbin=10, name="energy_true"
    )

    offset_axis = MapAxis.from_bounds(0, 1, nbin=4, name="offset", unit="deg")

    aeff = EffectiveAreaTable2D(
        axes=[energy_axis_true, offset_axis], data=1, unit="cm2"
    )
    hdu = aeff.to_table_hdu()
    assert_equal(hdu.data["ENERG_LO"][0], aeff.axes["energy_true"].edges[:-1].value)
    assert hdu.header["TUNIT1"] == aeff.axes["energy_true"].unit
    assert "FOVALIGN" not in hdu.header


def test_to_table_is_pointlike():
    energy_axis = MapAxis.from_energy_bounds(
        "1 TeV", "10 TeV", nbin=3, name="energy_true"
    )
    offset_axis = MapAxis.from_bounds(0 * u.deg, 2 * u.deg, nbin=2, name="offset")

    aeff = EffectiveAreaTable2D(
        data=np.ones((3, 2)) * u.m**2, axes=[energy_axis, offset_axis]
    )
    hdu = aeff.to_table_hdu()
    assert "is_pointlike" not in hdu.header


def test_wrong_axis_order():
    energy_axis_true = MapAxis.from_energy_bounds(
        "1 TeV", "10 TeV", nbin=10, name="energy_true"
    )

    offset = np.linspace(0, 1, 4) * u.deg
    offset_axis = MapAxis.from_nodes(offset, name="offset")

    data = np.ones(shape=(offset_axis.nbin, energy_axis_true.nbin))

    with pytest.raises(ValueError):
        EffectiveAreaTable2D(
            axes=[energy_axis_true, offset_axis], data=data, unit="cm2"
        )


def test_wrong_units():
    energy_axis_true = MapAxis.from_energy_bounds(
        "1 TeV", "10 TeV", nbin=10, name="energy_true"
    )

    offset_axis = MapAxis.from_bounds(0 * u.deg, 2 * u.deg, nbin=2, name="offset")

    wrong_unit = u.TeV
    data = np.ones((energy_axis_true.nbin, offset_axis.nbin)) * wrong_unit
    area_test = EffectiveAreaTable2D(axes=[energy_axis_true, offset_axis])

    with pytest.raises(ValueError) as error:
        EffectiveAreaTable2D(data=data, axes=[energy_axis_true, offset_axis])

        assert error.match(
            f"Error: {wrong_unit} is not an allowed unit. {area_test.tag} "
            f"requires {area_test.default_unit} data quantities."
        )


@requires_data("gammapy-data")
def test_aeff2d_pointlike():
    filename = "$GAMMAPY_DATA/joint-crab/dl3/magic/run_05029748_DL3.fits"

    aeff = EffectiveAreaTable2D.read(filename)
    hdu = aeff.to_table_hdu()

    assert aeff.is_pointlike
    assert hdu.header["HDUCLAS3"] == "POINT-LIKE"


def test_eq():
    energy1 = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=2, name="energy_true")

    energy2 = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=3, name="energy_true")

    offset_axis = MapAxis.from_bounds(0 * u.deg, 2 * u.deg, nbin=2, name="offset")

    data1 = np.ones((energy1.nbin, offset_axis.nbin)) * u.cm**2
    data2 = np.ones((energy2.nbin, offset_axis.nbin)) * u.cm**2

    aeff1 = EffectiveAreaTable2D(data=data1, axes=[energy1, offset_axis])
    aeff2 = EffectiveAreaTable2D(data=data2, axes=[energy2, offset_axis])

    assert not aeff1 == aeff2



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utility script to work with the CITATION.cff file"""
import logging
import subprocess
from pathlib import Path
import click
from ruamel.yaml import YAML

log = logging.getLogger(__name__)

EXCLUDE_AUTHORS = ["azure-pipelines[bot]", "GitHub Actions"]

GAMMAPY_CC = [
    "Axel Donath",
    "Bruno Khelifi",
    "Catherine Boisson",
    "Christopher van Eldik",
    "David Berge",
    "Fabio Acero",
    "Fabio Pintore",
    "James Hinton",
    "José Luis Contreras Gonzalez",
    "Matthias Fuessling",
    "Régis Terrier",
    "Roberta Zanin",
    "Rubén López-Coto",
    "Stefan Funk",
]

# Approved authors that requested to be added to CITATION.cff
ADDITIONAL_AUTHORS = [
    "Amanda Weinstein",
    "Tim Unbehaun",
]

PATH = Path(__file__).parent.parent

LAST_LTS = "v1.0"
NOW = "HEAD"


@click.group()
def cli():
    pass


def get_git_shortlog_authors(since_last_lts=False):
    """Get list of authors from git shortlog"""
    authors = []
    command = ("git", "shortlog", "--summary", "--numbered")

    if since_last_lts:
        command += (f"{LAST_LTS}..{NOW}",)

    result = subprocess.check_output(command, stderr=subprocess.STDOUT).decode()
    data = result.split("\n")
    for row in data:
        parts = row.split("\t")

        if len(parts) == 2:
            n_commits, author = parts
            if author not in EXCLUDE_AUTHORS:
                authors.append(author)

    return authors


def get_full_name(author_data):
    """Get full name from CITATION.cff parts"""
    parts = []
    parts.append(author_data["given-names"])

    name_particle = author_data.get("name-particle", None)

    if name_particle:
        parts.append(name_particle)

    parts.append(author_data["family-names"])
    return " ".join(parts)


def get_citation_cff_authors():
    """Get list of authors from CITATION.cff"""
    authors = []
    citation_file = PATH / "CITATION.cff"

    yaml = YAML()

    with citation_file.open("r") as stream:
        data = yaml.load(stream)

    for author_data in data["authors"]:
        full_name = get_full_name(author_data)
        authors.append(full_name)

    return authors


@cli.command("sort", help="Sort authors by commits")
def sort_citation_cff():
    """Sort CITATION.cff according to the git shortlog"""
    authors = get_git_shortlog_authors()
    citation_file = PATH / "CITATION.cff"

    yaml = YAML()
    yaml.preserve_quotes = True

    with citation_file.open("r") as stream:
        data = yaml.load(stream)

    authors_cff = get_citation_cff_authors()

    sorted_authors = []

    for author in authors:
        idx = authors_cff.index(author)
        sorted_authors.append(data["authors"][idx])

    data["authors"] = sorted_authors

    with citation_file.open("w") as stream:
        yaml.dump(data, stream=stream)


@cli.command("check", help="Check git shortlog vs CITATION.cff authors")
@click.option("--since-last-lts", is_flag=True, help="Show authors since last LTS")
def check_author_lists(since_last_lts):
    """Check CITATION.cff with git shortlog"""
    authors = set(get_git_shortlog_authors(since_last_lts))
    authors_cff = set(get_citation_cff_authors())

    message = "****Authors not in CITATION.cff****\n\n  "
    diff = authors.difference(authors_cff)
    print(message + "\n  ".join(sorted(diff)) + "\n")

    message = "****Authors not in shortlog****\n\n  "
    diff = authors_cff.difference(authors)

    authors_annotated = []

    for author in sorted(diff):
        if author in ADDITIONAL_AUTHORS:
            author = "(AA) " + author
        elif author in GAMMAPY_CC:
            author = "(CC) " + author
        else:
            author = 5 * " " + author

        authors_annotated.append(author)

    print(message + "\n  ".join(authors_annotated) + "\n")

    print("(CC) = Coordination Committe, (AA) = Additional Authors, (OO) = Opted out\n")


if __name__ == "__main__":
    cli()



"""
Account for spectral absorption due to the EBL
==============================================

Gamma rays emitted from extra-galactic objects, eg blazars, interact
with the photons of the Extragalactic Background Light (EBL) through
pair production and are attenuated, thus modifying the intrinsic
spectrum.

Various models of the EBL are supplied in `GAMMAPY_DATA`. This
notebook shows how to use these models to correct for this interaction.

"""

######################################################################
# Setup
# -----
#
# As usual, we’ll start with the standard imports …
#

import astropy.units as u
import matplotlib.pyplot as plt
from gammapy.catalog import SourceCatalog4FGL
from gammapy.datasets import SpectrumDatasetOnOff
from gammapy.estimators import FluxPointsEstimator
from gammapy.modeling import Fit
from gammapy.modeling.models import (
    EBL_DATA_BUILTIN,
    EBLAbsorptionNormSpectralModel,
    GaussianPrior,
    PowerLawSpectralModel,
    SkyModel,
)

######################################################################
# Load the data
# -------------
#
# We will use 6 observations of the blazars PKS 2155-304 taken in 2008 by
# H.E.S.S. when it was in a steady state. The data have already been
# reduced to OGIP format `~gammapy.datasets.SpectrumDatasetOnOff` following the procedure
# :doc:`/tutorials/analysis-1d/spectral_analysis` tutorial using a
# reflected regions background estimation. The spectra and IRFs from the
# 6 observations have been stacked together.
#
# We will load this dataset as a `~gammapy.datasets.SpectrumDatasetOnOff` and proceed with
# the modeling. You can do a 3D analysis as well.
#

dataset = SpectrumDatasetOnOff.read(
    "$GAMMAPY_DATA/PKS2155-steady/pks2155-304_steady.fits.gz"
)

print(dataset)


######################################################################
# Model the observed spectrum
# ---------------------------
#
# The observed spectrum is already attenuated due to the EBL. Assuming
# that the intrinsic spectrum is a power law, the observed spectrum is a
# `~gammapy.modeling.models.CompoundSpectralModel` given by the product of an
# EBL model with the intrinsic model.
#


######################################################################
# For a list of available models, see
# :doc:`/api/gammapy.modeling.models.EBL_DATA_BUILTIN`.
#

print(EBL_DATA_BUILTIN.keys())

######################################################################
# To use other EBL models, you need to save the optical depth as a
# function of energy and redshift as an XSPEC model.
# Alternatively, you can use packages like `ebltable <https://github.com/me-manu/ebltable>`_
# which shows how to interface other EBL models with Gammapy.
#

######################################################################
# Define the power law
#
index = 2.3
amplitude = 1.81 * 1e-12 * u.Unit("cm-2 s-1 TeV-1")
reference = 1 * u.TeV
pwl = PowerLawSpectralModel(index=index, amplitude=amplitude, reference=reference)
pwl.index.frozen = False
# Specify the redshift of the source
redshift = 0.116

# Load the EBL model. Here we use the model from Dominguez, 2011
absorption = EBLAbsorptionNormSpectralModel.read_builtin("dominguez", redshift=redshift)


# The power-law model is multiplied by the EBL to get the final model
spectral_model = pwl * absorption
print(spectral_model)

######################################################################
# Now, create a sky model and proceed with the fit
#
sky_model = SkyModel(spatial_model=None, spectral_model=spectral_model, name="pks2155")

dataset.models = sky_model

######################################################################
# Note that since this dataset has been produced
# by a reflected region analysis, it uses ON-OFF statistic
# and does not require a background model.
#

fit = Fit()
result = fit.run(datasets=[dataset])

# we make a copy here to compare it later
model_best = sky_model.copy()

print(result.models.to_parameters_table())


######################################################################
# Get the flux points
# ===================
#
# To get the observed flux points, just run the `~gammapy.estimators.FluxPointsEstimator`
# normally
#

energy_edges = dataset.counts.geom.axes["energy"].edges
fpe = FluxPointsEstimator(
    energy_edges=energy_edges, source="pks2155", selection_optional="all"
)
flux_points_obs = fpe.run(datasets=[dataset])


######################################################################
# To get the deabsorbed flux points (ie, intrinsic points), we simply need
# to set the reference model to the best fit power law instead of the
# compound model.
#

flux_points_intrinsic = flux_points_obs.copy(
    reference_model=SkyModel(spectral_model=pwl)
)

######################################################################
#
print(flux_points_obs.reference_model)

######################################################################
#
print(flux_points_intrinsic.reference_model)


######################################################################
# Plot the observed and intrinsic fluxes
# --------------------------------------
#

plt.figure()
sed_type = "e2dnde"
energy_bounds = [0.2, 20] * u.TeV
ax = flux_points_obs.plot(sed_type=sed_type, label="observed", color="navy")
flux_points_intrinsic.plot(ax=ax, sed_type=sed_type, label="intrinsic", color="red")

model_best.spectral_model.plot(
    ax=ax, energy_bounds=energy_bounds, sed_type=sed_type, color="blue"
)
model_best.spectral_model.plot_error(
    ax=ax, energy_bounds=energy_bounds, sed_type="e2dnde", facecolor="blue"
)

pwl.plot(ax=ax, energy_bounds=energy_bounds, sed_type=sed_type, color="tomato")
pwl.plot_error(
    ax=ax, energy_bounds=energy_bounds, sed_type=sed_type, facecolor="tomato"
)
plt.ylim(bottom=1e-13)
plt.legend()
plt.show()
# sphinx_gallery_thumbnail_number = 2


######################################################################
# Further extensions
# ------------------
#
# In this notebook, we have kept the parameters of the EBL model, the
# `alpha_norm` and the `redshift` frozen. Under reasonable assumptions
# on the intrinsic spectrum, it can be possible to constrain these
# parameters.
#
# Example: We now assume that the FermiLAT 4FGL catalog spectrum of the
# source is a good assumption of the intrinsic spectrum.
#
# *NOTE*: This is a very simplified assumption and in reality, EBL
# absorption can affect the Fermi spectrum significantly. Also, blazar
# spectra vary with time and long term averaged states may not be
# representative of a specific steady state
#

catalog = SourceCatalog4FGL()

src = catalog["PKS 2155-304"]

# Get the intrinsic model
intrinsic_model = src.spectral_model()
print(intrinsic_model)


######################################################################
# We add Gaussian priors on the `alpha` and `beta` parameters based on the 4FGL
# measurements and the associated errors. For more details on using priors, see
# :doc:`/tutorials/details/priors`
#

intrinsic_model.alpha.prior = GaussianPrior(
    mu=intrinsic_model.alpha.value, sigma=intrinsic_model.alpha.error
)
intrinsic_model.beta.prior = GaussianPrior(
    mu=intrinsic_model.beta.value, sigma=intrinsic_model.beta.error
)


######################################################################
# As before, multiply the intrinsic model with the EBL model
#

obs_model = intrinsic_model * absorption


######################################################################
# Now, free the redshift of the source
#

obs_model.parameters["redshift"].frozen = False

print(obs_model.parameters.to_table())

sky_model = SkyModel(spectral_model=obs_model, name="observed")
dataset.models = sky_model

result1 = fit.run([dataset])

print(result1.parameters.to_table())


######################################################################
# Get a fit stat profile for the redshift
# ---------------------------------------
#
# For more information about stat profiles, see
# :doc:`/tutorials/details/fitting`
#

total_stat = result1.total_stat

par = sky_model.parameters["redshift"]
par.scan_max = par.value + 5.0 * par.error
par.scan_min = max(0, par.value - 5.0 * par.error)
par.scan_n_values = 31

# %time
profile = fit.stat_profile(
    datasets=[dataset], parameter=sky_model.parameters["redshift"], reoptimize=True
)

plt.figure()
ax = plt.gca()
ax.plot(
    profile["observed.spectral.model2.redshift_scan"], profile["stat_scan"] - total_stat
)
ax.set_title("TS profile")
ax.set_xlabel("Redshift")
ax.set_ylabel("$\Delta$ TS")
plt.show()


######################################################################
# We see that the redshift is well constrained.
#



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.table import Column, Table
from astropy.time import Time
from astropy.timeseries import BinnedTimeSeries, BoxLeastSquares
from gammapy.data import GTI
from gammapy.datasets import Datasets, FluxPointsDataset
from gammapy.datasets.actors import DatasetsActor
from gammapy.estimators import FluxPoints, LightCurveEstimator
from gammapy.estimators.points.lightcurve import parallel as parallel
from gammapy.estimators.points.tests.test_sed import (
    simulate_map_dataset,
    simulate_spectrum_dataset,
)
from gammapy.modeling import Fit
from gammapy.modeling.models import FoVBackgroundModel, PowerLawSpectralModel, SkyModel
from gammapy.utils.testing import (
    assert_time_allclose,
    mpl_plot_check,
    requires_data,
    requires_dependency,
)


@pytest.fixture(scope="session")
def lc():
    meta = dict(TIMESYS="utc", SED_TYPE="flux")

    table = Table(
        meta=meta,
        data=[
            Column(Time(["2010-01-01", "2010-01-03"]).mjd, "time_min"),
            Column(Time(["2010-01-03", "2010-01-10"]).mjd, "time_max"),
            Column([[1.0], [1.0]], "e_min", unit="TeV"),
            Column([[2.0], [2.0]], "e_max", unit="TeV"),
            Column([1e-11, 3e-11], "flux", unit="cm-2 s-1"),
            Column([0.1e-11, 0.3e-11], "flux_err", unit="cm-2 s-1"),
            Column([np.nan, 3.6e-11], "flux_ul", unit="cm-2 s-1"),
            Column([False, True], "is_ul"),
            Column([True, True], "success"),
        ],
    )
    gti = GTI.create("0h", "1h", "2010-01-01T00:00:00")

    return FluxPoints.from_table(table=table, gti=gti)


@pytest.fixture(scope="session")
def lc_2d():
    meta = dict(TIMESYS="utc", SED_TYPE="flux")

    table = Table(
        meta=meta,
        data=[
            Column(Time(["2010-01-01", "2010-01-03"]).mjd, "time_min"),
            Column(Time(["2010-01-03", "2010-01-10"]).mjd, "time_max"),
            Column([[1.0, 2.0], [1.0, 2.0]], "e_min", unit="TeV"),
            Column([[2.0, 4.0], [2.0, 4.0]], "e_max", unit="TeV"),
            Column([[1e-11, 1e-12], [3e-11, 3e-12]], "flux", unit="cm-2 s-1"),
            Column([[0.1e-11, 1e-13], [0.3e-11, 3e-13]], "flux_err", unit="cm-2 s-1"),
            Column([[np.nan, np.nan], [3.6e-11, 3.6e-12]], "flux_ul", unit="cm-2 s-1"),
            Column([[False, False], [True, True]], "is_ul"),
        ],
    )

    return FluxPoints.from_table(table=table)


def test_lightcurve_str(lc):
    info_str = str(lc)
    assert "time" in info_str


def test_lightcurve_properties_time(lc):
    axis = lc.geom.axes["time"]

    assert axis.reference_time.scale == "utc"
    assert axis.reference_time.format == "mjd"

    # Time-related attributes
    time = axis.time_mid
    assert time.scale == "utc"
    assert time.format == "mjd"
    assert_allclose(time.mjd, [55198, 55202.5])

    assert_allclose(axis.time_min.mjd, [55197, 55199])
    assert_allclose(axis.time_max.mjd, [55199, 55206])

    # Note: I'm not sure why the time delta has this scale and format
    time_delta = axis.time_delta
    assert time_delta.scale == "tai"
    assert time_delta.format == "jd"
    assert_allclose(time_delta.jd, [2, 7])


def test_lightcurve_properties_flux(lc):
    table = lc.to_table(sed_type="flux")
    flux = table["flux"].quantity
    assert flux.unit == "cm-2 s-1"
    assert_allclose(flux.value, [[1e-11], [3e-11]])


# TODO: extend these tests to cover other time scales.
# In those cases, CSV should not round-trip because there
# is no header info in CSV to store the time scale!


@pytest.mark.parametrize("sed_type", ["dnde", "flux", "likelihood"])
def test_lightcurve_read_write(tmp_path, lc, sed_type):
    lc.write(tmp_path / "tmp.fits", sed_type=sed_type)

    lc = FluxPoints.read(tmp_path / "tmp.fits")
    assert_allclose(lc.gti.time_start[0].mjd, 55197.000766, rtol=1e-5)

    # Check if time-related info round-trips
    axis = lc.geom.axes["time"]
    assert axis.reference_time.scale == "utc"
    assert axis.reference_time.format == "mjd"
    assert_allclose(axis.time_mid.mjd, [55198, 55202.5])


def test_lightcurve_plot(lc, lc_2d):
    with mpl_plot_check():
        lc.plot()

    with mpl_plot_check():
        lc.plot(marker="o", time_format="mjd")

    with mpl_plot_check():
        lc_2d.plot(axis_name="time")


@requires_data()
def test_lightcurve_to_time_series():
    from gammapy.catalog import SourceCatalog4FGL

    catalog_4fgl = SourceCatalog4FGL("$GAMMAPY_DATA/catalogs/fermi/gll_psc_v20.fit.gz")
    lightcurve = catalog_4fgl["FGES J1553.8-5325"].lightcurve()

    table = lightcurve.to_table(sed_type="flux", format="binned-time-series")

    timeseries = BinnedTimeSeries(data=table)

    assert_allclose(timeseries.time_bin_center.mjd[0], 54863.97885336907)
    assert_allclose(timeseries.time_bin_end.mjd[0], 55045.301668796295)

    time_axis = lightcurve.geom.axes["time"]
    assert_allclose(timeseries.time_bin_end.mjd[0], time_axis.time_max.mjd[0])

    # assert that it interfaces with periodograms

    p = BoxLeastSquares.from_timeseries(
        timeseries=timeseries, signal_column_name="flux", uncertainty="flux_errp"
    )

    result = p.power(1 * u.year, 0.5 * u.year)
    assert_allclose(result.duration, 182.625 * u.d)


def get_spectrum_datasets():
    model = SkyModel(spectral_model=PowerLawSpectralModel())
    dataset_1 = simulate_spectrum_dataset(model=model, random_state=0)
    dataset_1._name = "dataset_1"
    gti1 = GTI.create("0h", "1h", Time("2010-01-01T00:00:00").tt)
    dataset_1.gti = gti1

    dataset_2 = simulate_spectrum_dataset(model=model, random_state=1)
    dataset_2._name = "dataset_2"
    gti2 = GTI.create("1h", "2h", Time("2010-01-01T00:00:00").tt)
    dataset_2.gti = gti2

    return [dataset_1, dataset_2]


@requires_data()
def test_group_datasets_in_time_interval():
    # Doing a LC on one hour bin
    datasets = Datasets(get_spectrum_datasets())
    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T01:00:00"]),
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]),
    ]

    group_table = datasets.gti.group_table(time_intervals)

    assert len(group_table) == 2
    assert_allclose(group_table["time_min"], [55197.0, 55197.04166666667])
    assert_allclose(group_table["time_max"], [55197.04166666667, 55197.083333333336])
    assert_allclose(group_table["group_idx"], [0, 1])


@requires_data()
def test_group_datasets_in_time_interval_outflows():
    datasets = Datasets(get_spectrum_datasets())
    # Check Overflow
    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T00:55:00"]),
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]),
    ]

    group_table = datasets.gti.group_table(time_intervals)
    assert group_table["bin_type"][0] == "overflow"

    # Check underflow
    time_intervals = [
        Time(["2010-01-01T00:05:00", "2010-01-01T01:00:00"]),
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]),
    ]
    group_table = datasets.gti.group_table(time_intervals)
    assert group_table["bin_type"][0] == "underflow"


@requires_data()
def test_lightcurve_estimator_fit_options():
    # Doing a LC on one hour bin
    datasets = get_spectrum_datasets()
    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T01:00:00"]),
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]),
    ]
    energy_edges = [1, 30] * u.TeV

    estimator = LightCurveEstimator(
        energy_edges=energy_edges,
        time_intervals=time_intervals,
        selection_optional="all",
        fit=Fit(backend="minuit", optimize_opts=dict(tol=0.2, strategy=1)),
        norm=dict(scan_n_values=3),
    )

    assert_allclose(estimator.fit.optimize_opts["tol"], 0.2)

    result = estimator.fit.run(datasets=datasets)
    assert_allclose(result.minuit.tol, 0.2)


@requires_data()
def test_lightcurve_estimator_spectrum_datasets():
    # Doing a LC on one hour bin
    datasets = get_spectrum_datasets()
    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T01:00:00"]).tt,
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]).tt,
    ]

    estimator = LightCurveEstimator(
        energy_edges=[1, 30] * u.TeV,
        time_intervals=time_intervals,
        selection_optional="all",
    )
    estimator.norm.scan_n_values = 3

    lightcurve = estimator.run(datasets)
    table = lightcurve.to_table()
    assert_allclose(table["time_min"], [55197.0, 55197.041667])
    assert_allclose(table["time_max"], [55197.041667, 55197.083333])
    assert_allclose(table["e_ref"], [[5.623413], [5.623413]])
    assert_allclose(table["e_min"], [[1], [1]])
    assert_allclose(table["e_max"], [[31.622777], [31.622777]])
    assert_allclose(table["ref_dnde"], [[3.162278e-14], [3.162278e-14]], rtol=1e-5)
    assert_allclose(table["ref_flux"], [[9.683772e-13], [9.683772e-13]], rtol=1e-5)
    assert_allclose(table["ref_eflux"], [[3.453878e-12], [3.453878e-12]], rtol=1e-5)
    assert_allclose(table["stat"], [[16.824042], [17.391981]], rtol=1e-5)
    assert_allclose(table["norm"], [[0.911963], [0.9069318]], rtol=1e-2)
    assert_allclose(table["norm_err"], [[0.057769], [0.057835]], rtol=1e-2)
    assert_allclose(table["counts"], [[[791, np.nan]], [[np.nan, 784]]])
    assert_allclose(table["norm_errp"], [[0.058398], [0.058416]], rtol=1e-2)
    assert_allclose(table["norm_errn"], [[0.057144], [0.057259]], rtol=1e-2)
    assert_allclose(table["norm_ul"], [[1.029989], [1.025061]], rtol=1e-2)
    assert_allclose(table["sqrt_ts"], [[19.384781], [19.161769]], rtol=1e-2)
    assert_allclose(table["ts"], [[375.769735], [367.173374]], rtol=1e-2)
    assert_allclose(table[0]["norm_scan"], [[0.2, 1.0, 5.0]])
    assert_allclose(
        table[0]["stat_scan"],
        [[224.058304, 19.074405, 2063.75636]],
        rtol=1e-5,
    )
    assert table.meta["TIMESYS"] == "tt"
    assert table.meta["TIMEUNIT"] == "d"
    assert table.meta["MJDREFF"] == 0
    assert table.meta["MJDREFI"] == 0

    # TODO: fix reference model I/O
    fp = FluxPoints.from_table(table=table, reference_model=PowerLawSpectralModel())
    assert fp.norm.geom.axes.names == ["energy", "time"]
    assert fp.counts.geom.axes.names == ["dataset", "energy", "time"]
    assert fp.stat_scan.geom.axes.names == ["norm", "energy", "time"]
    assert_time_allclose(
        fp.geom.axes["time"].time_min, lightcurve.geom.axes["time"].time_min
    )
    assert_time_allclose(
        fp.geom.axes["time"].time_max, lightcurve.geom.axes["time"].time_max
    )


@requires_data()
def test_lightcurve_estimator_spectrum_datasets_2_energy_bins():
    # Doing a LC on one hour bin
    datasets = get_spectrum_datasets()
    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T01:00:00"]),
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]),
    ]

    estimator = LightCurveEstimator(
        energy_edges=[1, 5, 30] * u.TeV,
        time_intervals=time_intervals,
        selection_optional="all",
    )
    estimator.norm.scan_n_values = 3
    lightcurve = estimator.run(datasets)
    table = lightcurve.to_table()

    assert_allclose(table["time_min"], [55197.0, 55197.041667])
    assert_allclose(table["time_max"], [55197.041667, 55197.083333])
    assert_allclose(table["e_ref"], [[2.238721, 12.589254], [2.238721, 12.589254]])
    assert_allclose(table["e_min"], [[1, 5.011872], [1, 5.011872]])
    assert_allclose(table["e_max"], [[5.011872, 31.622777], [5.011872, 31.622777]])
    assert_allclose(
        table["ref_dnde"],
        [[1.995262e-13, 6.309573e-15], [1.995262e-13, 6.309573e-15]],
        rtol=1e-5,
    )
    assert_allclose(
        table["stat"],
        [[8.234951, 8.30321], [2.037205, 15.300507]],
        rtol=1e-5,
    )
    assert_allclose(
        table["norm"],
        [[0.894723, 0.967419], [0.914283, 0.882351]],
        rtol=1e-2,
    )
    assert_allclose(
        table["norm_err"],
        [[0.065905, 0.121288], [0.06601, 0.119457]],
        rtol=1e-2,
    )
    assert_allclose(
        table["counts"],
        [[[669.0, np.nan], [122.0, np.nan]], [[np.nan, 667.0], [np.nan, 117.0]]],
    )
    assert_allclose(
        table["norm_errp"],
        [[0.06664, 0.124741], [0.066815, 0.122832]],
        rtol=1e-2,
    )
    assert_allclose(
        table["norm_errn"],
        [[0.065176, 0.117904], [0.065212, 0.116169]],
        rtol=1e-2,
    )
    assert_allclose(
        table["norm_ul"],
        [[1.029476, 1.224117], [1.049283, 1.134874]],
        rtol=1e-2,
    )
    assert_allclose(
        table["sqrt_ts"],
        [[16.233236, 10.608376], [16.609784, 9.557339]],
        rtol=1e-2,
    )
    assert_allclose(table[0]["norm_scan"], [[0.2, 1.0, 5.0], [0.2, 1.0, 5.0]])
    assert_allclose(
        table[0]["stat_scan"],
        [[153.880281, 10.701492, 1649.609684], [70.178023, 8.372913, 414.146676]],
        rtol=1e-5,
    )

    # those quantities are currently not part of the table so we test separately
    npred = lightcurve.npred.data.squeeze()
    assert_allclose(
        npred,
        [[[669.36, np.nan], [121.66, np.nan]], [[np.nan, 664.41], [np.nan, 115.09]]],
        rtol=1e-3,
    )

    npred_excess_err = lightcurve.npred_excess_err.data.squeeze()
    assert_allclose(
        npred_excess_err,
        [[[26.80, np.nan], [11.31, np.nan]], [[np.nan, 26.85], [np.nan, 11.14]]],
        rtol=1e-3,
    )

    npred_excess_errp = lightcurve.npred_excess_errp.data.squeeze()
    assert_allclose(
        npred_excess_errp,
        [[[27.11, np.nan], [11.63, np.nan]], [[np.nan, 27.15], [np.nan, 11.46]]],
        rtol=1e-3,
    )

    npred_excess_errn = lightcurve.npred_excess_errn.data.squeeze()
    assert_allclose(
        npred_excess_errn,
        [[[26.50, np.nan], [11.00, np.nan]], [[np.nan, 26.54], [np.nan, 10.84]]],
        rtol=1e-3,
    )

    npred_excess_ul = lightcurve.npred_excess_ul.data.squeeze()
    assert_allclose(
        npred_excess_ul,
        [[[418.68, np.nan], [114.19, np.nan]], [[np.nan, 426.74], [np.nan, 105.86]]],
        rtol=1e-3,
    )

    fp = FluxPoints.from_table(table=table, reference_model=PowerLawSpectralModel())
    assert fp.norm.geom.axes.names == ["energy", "time"]
    assert fp.counts.geom.axes.names == ["dataset", "energy", "time"]
    assert fp.stat_scan.geom.axes.names == ["norm", "energy", "time"]


@requires_data()
def test_lightcurve_estimator_spectrum_datasets_with_mask_fit():
    # Doing a LC on one hour bin
    datasets = get_spectrum_datasets()
    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T01:00:00"]),
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]),
    ]

    energy_min_fit = 1 * u.TeV
    energy_max_fit = 3 * u.TeV
    for dataset in datasets:
        geom = dataset.counts.geom
        dataset.mask_fit = geom.energy_mask(
            energy_min=energy_min_fit, energy_max=energy_max_fit
        )

    selection = ["scan"]
    estimator = LightCurveEstimator(
        energy_edges=[1, 30] * u.TeV,
        time_intervals=time_intervals,
        selection_optional=selection,
    )
    estimator.norm.scan_n_values = 3
    lightcurve = estimator.run(datasets)
    table = lightcurve.to_table()
    assert_allclose(table["time_min"], [55197.0, 55197.041667])
    assert_allclose(table["time_max"], [55197.041667, 55197.083333])
    assert_allclose(table["stat"], [[6.603043], [0.421051]], rtol=1e-3)
    assert_allclose(table["norm"], [[0.885124], [0.967054]], rtol=1e-3)


@requires_data()
def test_lightcurve_estimator_spectrum_datasets_default():
    # Test default time interval: each time interval is equal to the gti of each
    # dataset, here one hour
    datasets = get_spectrum_datasets()
    selection = ["scan"]
    estimator = LightCurveEstimator(
        energy_edges=[1, 30] * u.TeV, selection_optional=selection
    )
    estimator.norm.scan_n_values = 3
    lightcurve = estimator.run(datasets)
    table = lightcurve.to_table()
    assert_allclose(table["time_min"], [55197.0, 55197.041667])
    assert_allclose(table["time_max"], [55197.041667, 55197.083333])
    assert_allclose(table["norm"], [[0.911963], [0.906931]], rtol=1e-3)


@requires_data()
def test_lightcurve_estimator_spectrum_datasets_notordered():
    # Test that if the time intervals given are not ordered in time, it is first ordered
    # correctly and then compute as expected
    datasets = get_spectrum_datasets()
    time_intervals = [
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]),
        Time(["2010-01-01T00:00:00", "2010-01-01T01:00:00"]),
    ]
    estimator = LightCurveEstimator(
        energy_edges=[1, 100] * u.TeV,
        time_intervals=time_intervals,
        selection_optional=["scan"],
    )
    estimator.norm.scan_n_values = 3
    lightcurve = estimator.run(datasets)
    table = lightcurve.to_table()
    assert_allclose(table["time_min"], [55197.0, 55197.041667])
    assert_allclose(table["time_max"], [55197.041667, 55197.083333])
    assert_allclose(table["norm"], [[0.911963], [0.906931]], rtol=1e-3)


@requires_data()
def test_lightcurve_estimator_spectrum_datasets_largerbin():
    # Test all dataset in a single LC bin, here two hours
    datasets = get_spectrum_datasets()
    time_intervals = [Time(["2010-01-01T00:00:00", "2010-01-01T02:00:00"])]
    estimator = LightCurveEstimator(
        energy_edges=[1, 30] * u.TeV,
        time_intervals=time_intervals,
        selection_optional=["scan"],
    )
    estimator.norm.scan_n_values = 3
    lightcurve = estimator.run(datasets)
    table = lightcurve.to_table()

    assert_allclose(table["time_min"], [55197.0])
    assert_allclose(table["time_max"], [55197.083333])
    assert_allclose(table["e_ref"][0], [5.623413])
    assert_allclose(table["e_min"][0], [1])
    assert_allclose(table["e_max"][0], [31.622777])
    assert_allclose(table["ref_dnde"][0], [3.162278e-14], rtol=1e-5)
    assert_allclose(table["ref_flux"][0], [9.683772e-13], rtol=1e-5)
    assert_allclose(table["ref_eflux"][0], [3.453878e-12], rtol=1e-5)
    assert_allclose(table["stat"][0], [34.219808], rtol=1e-5)
    assert_allclose(table["norm"][0], [0.909646], rtol=1e-5)
    assert_allclose(table["norm_err"][0], [0.040874], rtol=1e-3)
    assert_allclose(table["ts"][0], [742.939324], rtol=1e-4)


@requires_data()
def test_lightcurve_estimator_spectrum_datasets_emptybin():
    # Test all dataset in a single LC bin, here two hours
    datasets = get_spectrum_datasets()
    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T02:00:00"]),
        Time(["2010-02-01T00:00:00", "2010-02-01T02:00:00"]),
    ]
    estimator = LightCurveEstimator(
        energy_edges=[1, 30] * u.TeV,
        time_intervals=time_intervals,
        selection_optional=["scan"],
    )
    estimator.norm.scan_n_values = 3
    lightcurve = estimator.run(datasets)
    table = lightcurve.to_table()

    assert_allclose(table["time_min"], [55197.0])
    assert_allclose(table["time_max"], [55197.083333])


@requires_data()
def test_lightcurve_estimator_spectrum_datasets_timeoverlaped():
    # Check that it returns a ValueError if the time intervals overlapped
    datasets = get_spectrum_datasets()
    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T01:30:00"]),
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]),
    ]
    with pytest.raises(ValueError) as excinfo:
        estimator = LightCurveEstimator(time_intervals=time_intervals)
        estimator.norm.scan_n_values = 3
        estimator.run(datasets)

    msg = "Overlapping time bins"
    assert str(excinfo.value) == msg


@requires_data()
def test_lightcurve_estimator_spectrum_datasets_gti_not_include_in_time_intervals():
    # Check that it returns a ValueError if the time intervals are smaller than the dataset GTI.
    datasets = get_spectrum_datasets()
    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T00:05:00"]),
        Time(["2010-01-01T01:00:00", "2010-01-01T01:05:00"]),
    ]
    estimator = LightCurveEstimator(
        energy_edges=[1, 30] * u.TeV,
        time_intervals=time_intervals,
        selection_optional=["scan"],
    )
    estimator.norm.scan_n_values = 3
    with pytest.raises(ValueError) as excinfo:
        estimator.run(datasets)
    msg = "LightCurveEstimator: No datasets in time intervals"
    assert str(excinfo.value) == msg


def get_map_datasets():
    dataset_1 = simulate_map_dataset(random_state=0, name="dataset_1")
    gti1 = GTI.create("0 h", "1 h", "2010-01-01T00:00:00")
    dataset_1.gti = gti1

    dataset_2 = simulate_map_dataset(random_state=1, name="dataset_2")
    gti2 = GTI.create("1 h", "2 h", "2010-01-01T00:00:00")
    dataset_2.gti = gti2

    model = dataset_1.models["source"].copy(name="test_source")
    bkg_model_1 = FoVBackgroundModel(dataset_name="dataset_1")
    bkg_model_2 = FoVBackgroundModel(dataset_name="dataset_2")

    dataset_1.models = [model, bkg_model_1]
    dataset_2.models = [model, bkg_model_2]

    return [dataset_1, dataset_2]


@requires_data()
def test_lightcurve_estimator_map_datasets():
    datasets = get_map_datasets()

    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T01:00:00"]),
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]),
    ]
    estimator = LightCurveEstimator(
        energy_edges=[1, 100] * u.TeV,
        source="test_source",
        time_intervals=time_intervals,
        selection_optional=["scan"],
    )
    lightcurve = estimator.run(datasets)
    table = lightcurve.to_table()
    assert_allclose(table["time_min"], [55197.0, 55197.041667])
    assert_allclose(table["time_max"], [55197.041667, 55197.083333])
    assert_allclose(table["e_ref"], [[10.857111], [10.857111]])
    assert_allclose(table["e_min"], [[1.178769], [1.178769]], rtol=1e-5)
    assert_allclose(table["e_max"], [[100], [100]])
    assert_allclose(table["ref_dnde"], [[8.483429e-14], [8.483429e-14]], rtol=1e-5)
    assert_allclose(table["ref_flux"], [[8.383429e-12], [8.383429e-12]], rtol=1e-5)
    assert_allclose(table["ref_eflux"], [[4.4407e-11], [4.4407e-11]], rtol=1e-5)
    assert_allclose(table["stat"], [[9402.778975], [9517.750207]], rtol=1e-2)
    assert_allclose(table["norm"], [[0.971592], [0.963286]], rtol=1e-2)
    assert_allclose(table["norm_err"], [[0.044643], [0.044475]], rtol=1e-2)
    assert_allclose(table["sqrt_ts"], [[35.880361], [35.636547]], rtol=1e-2)
    assert_allclose(table["ts"], [[1287.4003], [1269.963491]], rtol=1e-2)

    datasets = get_map_datasets()

    time_intervals2 = [Time(["2010-01-01T00:00:00", "2010-01-01T02:00:00"])]
    estimator2 = LightCurveEstimator(
        energy_edges=[1, 100] * u.TeV,
        source="test_source",
        time_intervals=time_intervals2,
        selection_optional=["scan"],
    )
    lightcurve2 = estimator2.run(datasets)
    table = lightcurve2.to_table()
    assert_allclose(table["time_min"][0], [55197.0])
    assert_allclose(table["time_max"][0], [55197.083333])
    assert_allclose(table["e_ref"][0], [10.857111], rtol=1e-5)
    assert_allclose(table["e_min"][0], [1.178769], rtol=1e-5)
    assert_allclose(table["e_max"][0], [100])
    assert_allclose(table["ref_dnde"][0], [8.483429e-14], rtol=1e-5)
    assert_allclose(table["ref_flux"][0], [8.383429e-12], rtol=1e-5)
    assert_allclose(table["ref_eflux"][0], [4.4407e-11], rtol=1e-5)
    assert_allclose(table["stat"][0], [18920.54651], rtol=1e-2)
    assert_allclose(table["norm"][0], [0.967438], rtol=1e-2)
    assert_allclose(table["norm_err"][0], [0.031508], rtol=1e-2)
    assert_allclose(table["counts"][0], [[2205, 2220]])
    assert_allclose(table["ts"][0], [2557.346464], rtol=1e-2)


@requires_data()
@requires_dependency("ray")
def test_lightcurve_estimator_map_datasets_ray_actors():
    datasets = get_map_datasets()

    datasets = DatasetsActor(datasets)

    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T01:00:00"]),
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]),
    ]
    estimator = LightCurveEstimator(
        energy_edges=[1, 100] * u.TeV,
        source="test_source",
        time_intervals=time_intervals,
        selection_optional=["scan"],
    )
    lightcurve = estimator.run(datasets)
    table = lightcurve.to_table()
    assert_allclose(table["time_min"], [55197.0, 55197.041667])
    assert_allclose(table["time_max"], [55197.041667, 55197.083333])
    assert_allclose(table["e_ref"], [[10.857111], [10.857111]])
    assert_allclose(table["e_min"], [[1.178769], [1.178769]], rtol=1e-5)
    assert_allclose(table["e_max"], [[100], [100]])
    assert_allclose(table["ref_dnde"], [[8.483429e-14], [8.483429e-14]], rtol=1e-5)
    assert_allclose(table["ref_flux"], [[8.383429e-12], [8.383429e-12]], rtol=1e-5)
    assert_allclose(table["ref_eflux"], [[4.4407e-11], [4.4407e-11]], rtol=1e-5)
    assert_allclose(table["stat"], [[9402.778975], [9517.750207]], rtol=1e-2)
    assert_allclose(table["norm"], [[0.971592], [0.963286]], rtol=1e-2)
    assert_allclose(table["norm_err"], [[0.044643], [0.044475]], rtol=1e-2)
    assert_allclose(table["sqrt_ts"], [[35.880361], [35.636547]], rtol=1e-2)
    assert_allclose(table["ts"], [[1287.4003], [1269.963491]], rtol=1e-2)

    datasets = get_map_datasets()

    time_intervals2 = [Time(["2010-01-01T00:00:00", "2010-01-01T02:00:00"])]
    estimator2 = LightCurveEstimator(
        energy_edges=[1, 100] * u.TeV,
        source="test_source",
        time_intervals=time_intervals2,
        selection_optional=["scan"],
    )
    lightcurve2 = estimator2.run(datasets)
    table = lightcurve2.to_table()
    assert_allclose(table["time_min"][0], [55197.0])
    assert_allclose(table["time_max"][0], [55197.083333])
    assert_allclose(table["e_ref"][0], [10.857111], rtol=1e-5)
    assert_allclose(table["e_min"][0], [1.178769], rtol=1e-5)
    assert_allclose(table["e_max"][0], [100])
    assert_allclose(table["ref_dnde"][0], [8.483429e-14], rtol=1e-5)
    assert_allclose(table["ref_flux"][0], [8.383429e-12], rtol=1e-5)
    assert_allclose(table["ref_eflux"][0], [4.4407e-11], rtol=1e-5)
    assert_allclose(table["stat"][0], [18920.54651], rtol=1e-2)
    assert_allclose(table["norm"][0], [0.967438], rtol=1e-2)
    assert_allclose(table["norm_err"][0], [0.031508], rtol=1e-2)
    assert_allclose(table["counts"][0], [[2205, 2220]])
    assert_allclose(table["ts"][0], [2557.346464], rtol=1e-2)


@requires_data()
def test_recompute_ul():
    datasets = get_spectrum_datasets()
    selection = ["all"]
    estimator = LightCurveEstimator(
        energy_edges=[1, 3, 10, 30] * u.TeV, selection_optional=selection, n_sigma_ul=2
    )
    lightcurve = estimator.run(datasets)

    assert_allclose(
        lightcurve.dnde_ul.data[0],
        [[[3.26070325e-13]], [[3.82484628e-14]], [[4.16433528e-15]]],
        rtol=1e-3,
    )

    new_lightcurve = lightcurve.recompute_ul(n_sigma_ul=4)

    assert_allclose(
        new_lightcurve.dnde_ul.data[0],
        [[[3.75813611e-13]], [[4.63590079e-14]], [[5.64634904e-15]]],
        rtol=1e-3,
    )

    assert new_lightcurve.meta["n_sigma_ul"] == 4

    # test if scan is not present
    selection = []
    estimator = LightCurveEstimator(
        energy_edges=[1, 30] * u.TeV, selection_optional=selection
    )
    lightcurve = estimator.run(datasets)
    with pytest.raises(ValueError):
        lightcurve.recompute_ul(n_sigma_ul=4)


@requires_data()
def test_dataset_stat_type():
    datasets = get_spectrum_datasets()
    selection = ["all"]
    estimator = LightCurveEstimator(
        energy_edges=[1, 3, 30] * u.TeV, selection_optional=selection, n_sigma_ul=2
    )
    lightcurve = estimator.run(datasets)
    assert_allclose(
        lightcurve.dnde_ul.data[0], [[[3.260703e-13]], [[1.159354e-14]]], rtol=1e-3
    )

    lightcurve_dataset = FluxPointsDataset(
        data=lightcurve, models=[lightcurve.reference_model]
    )
    lightcurve_dataset.stat_type = "profile"
    assert_allclose(lightcurve_dataset.stat_sum(), 5.741539, rtol=1e-5)
    lightcurve_dataset.stat_type = "distrib"
    assert_allclose(lightcurve_dataset.stat_sum(), 5.824218, rtol=1e-5)
    lightcurve_dataset.stat_type = "chi2"
    assert_allclose(lightcurve_dataset.stat_sum(), 6.014128, rtol=1e-5)

    # test if scan is not present
    selection = []
    estimator = LightCurveEstimator(
        energy_edges=[1, 30] * u.TeV, selection_optional=selection
    )
    lightcurve = estimator.run(datasets)

    lightcurve_dataset = FluxPointsDataset(
        data=lightcurve, models=[lightcurve.reference_model]
    )
    with pytest.raises(AttributeError):
        lightcurve_dataset.stat_type = "profile"
        lightcurve_dataset.stat_sum()


@requires_data()
def test_lightcurve_parallel_multiprocessing():
    datasets = get_spectrum_datasets()
    selection = ["all"]
    estimator = LightCurveEstimator(
        energy_edges=[1, 3, 30] * u.TeV,
        selection_optional=selection,
        n_sigma_ul=2,
        n_jobs=2,
    )
    assert estimator.n_jobs == 2
    lightcurve = estimator.run(datasets)
    assert_allclose(
        lightcurve.dnde_ul.data[0], [[[3.260703e-13]], [[1.159354e-14]]], rtol=1e-3
    )
    assert estimator._get_n_child_jobs == 1

    parallel.ALLOW_CHILD_JOBS = True
    lightcurve = estimator.run(datasets)
    assert_allclose(
        lightcurve.dnde_ul.data[0], [[[3.260703e-13]], [[1.159354e-14]]], rtol=1e-3
    )
    assert estimator._get_n_child_jobs == 2
    parallel.ALLOW_CHILD_JOBS = False


@requires_data()
@requires_dependency("ray")
def test_lightcurve_parallel_ray():
    datasets = get_spectrum_datasets()
    selection = ["all"]

    estimator = LightCurveEstimator(
        energy_edges=[1, 3, 30] * u.TeV,
        selection_optional=selection,
        n_sigma_ul=2,
        n_jobs=2,
        parallel_backend="ray",
    )

    assert estimator.n_jobs == 2
    lightcurve = estimator.run(datasets)
    assert_allclose(
        lightcurve.dnde_ul.data[0], [[[3.260703e-13]], [[1.159354e-14]]], rtol=1e-3
    )
    assert estimator._get_n_child_jobs == 1

    parallel.ALLOW_CHILD_JOBS = True
    lightcurve = estimator.run(datasets)
    assert_allclose(
        lightcurve.dnde_ul.data[0], [[[3.260703e-13]], [[1.159354e-14]]], rtol=1e-3
    )
    assert estimator._get_n_child_jobs == 2
    parallel.ALLOW_CHILD_JOBS = False


@requires_data()
def test_lightcurve_stacking():
    datasets = get_spectrum_datasets()
    time_intervals = [
        Time(["2010-01-01T00:00:00", "2010-01-01T01:00:00"]),
        Time(["2010-01-01T01:00:00", "2010-01-01T02:00:00"]),
    ]
    energy_edges = [1, 10] * u.TeV

    # Internal stacking
    estimator_stack = LightCurveEstimator(
        energy_edges=energy_edges,
        time_intervals=time_intervals,
        stack_over_time_interval=True,
    )
    results_stack = estimator_stack.run(datasets)
    table_stack = results_stack.to_table()

    # External stacking
    datasets = Datasets(datasets)
    models = datasets.models
    stacked_datasets = Datasets()

    for t_min, t_max in time_intervals:
        selected = datasets.select_time(time_min=t_min, time_max=t_max, atol="1e-6 s")
        stacked = selected.stack_reduce()
        stacked_datasets.append(stacked)

    stacked_datasets.models = models
    estimator = LightCurveEstimator(
        energy_edges=energy_edges,
        time_intervals=time_intervals,
        stack_over_time_interval=False,
    )
    results_stacked = estimator.run(stacked_datasets)
    table_stack_outside = results_stacked.to_table()
    assert_allclose(table_stack["stat_null"], table_stack_outside["stat_null"])
